"""Contains the Timed Switch device class."""
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.mode_device import ModeDevice
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor("active_switches")
class TimedSwitch(SystemWideDevice, ModeDevice):

    """Timed Switch device."""

    config_section = 'timed_switches'
    collection = 'timed_switches'
    class_label = 'timed_switch'

    def __init__(self, machine, name):
        """Initialize Timed Switch."""
        super().__init__(machine, name)
        self.active_switches = set()

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Validate and parse config."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)

        for x in ('active', 'released'):
            if not config['events_when_{}'.format(x)]:
                config['events_when_{}'.format(x)] = [
                    "{}_{}".format(self.name, x)]

        if config['state'] == 'active':
            config['state'] = 1
        elif config['state'] == 'inactive':
            config['state'] = 0

        return config

    async def _initialize(self):
        await super()._initialize()

        for tag in self.config['switch_tags']:
            for switch in self.machine.switches.items_tagged(tag):
                if switch not in self.config['switches']:
                    self.config['switches'].append(switch)

        self._register_switch_handlers()

    @property
    def can_exist_outside_of_game(self):
        """Return true if this device can exist outside of a game."""
        return True

    def device_removed_from_mode(self, mode):
        """Mode ended.

        Args:
        ----
            mode: mode which stopped
        """
        del mode
        self._remove_switch_handlers()

    def _register_switch_handlers(self):
        for switch in self.config['switches']:
            switch.add_handler(self._activate,
                               state=self.config['state'] ^ 0,
                               ms=self.config['time'], return_info=True)
            switch.add_handler(self._deactivate,
                               state=self.config['state'] ^ 1,
                               ms=0, return_info=True)

    def _remove_switch_handlers(self):
        for switch in self.config['switches']:
            switch.remove_handler(self._activate,
                                  state=1 if self.config['state'] else 0)
            switch.remove_handler(self._deactivate,
                                  state=0 if self.config['state'] else 1)

    def _activate(self, switch_name, state, ms):
        del state, ms
        if not self.active_switches:
            for event in self.config['events_when_active']:
                self.machine.events.post(event)

        self.active_switches.add(switch_name)

    def _deactivate(self, switch_name, state, ms):
        del state, ms

        try:

            # why 'try' with 'remove' instead of 'discard'? Because we only
            # want to post the event if it was active and there is no more
            # active

            self.active_switches.remove(switch_name)

            if not self.active_switches:
                for event in self.config['events_when_released']:
                    self.machine.events.post(event)
        except KeyError:
            pass

        '''event: (name)_active
        config_attribute: events_when_active

        desc: Posted when one of the switches buttons has been active for ``time``.
        '''

        '''event: (name)_released
        config_attribute: events_when_released

        desc: Posted when one of the switches that has previously
        been active for more than ``time`` has been released.
        '''



"""I2C servo controller platform."""
import asyncio
import logging

from mpf.exceptions.config_file_error import ConfigFileError
from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface

from mpf.core.platform import ServoPlatform


class I2CServoControllerHardwarePlatform(ServoPlatform):

    """Supports the PCA9685/PCA9635 chip via I2C."""

    def __init__(self, machine):
        """Initialise I2C servo platform."""
        super().__init__(machine)
        self.config = self.machine.config_validator.validate_config("servo_controllers",
                                                                    self.machine.config.get('servo_controllers', {}))
        self.platform = None
        self.i2c_devices = {}
        self.features['tickless'] = True
        self._configure_device_logging_and_debug("I2C Servo Controller", self.config)

    def __repr__(self):
        """Return string representation."""
        return '<Platform.I2C_Servo_Controller_Platform>'

    async def initialize(self):
        """Initialise platform."""
        await super().initialize()
        # load i2c platform
        self.platform = self.machine.get_platform_sections(
            "i2c", self.config['platform'])
        self.platform.assert_has_feature("i2c")

    async def _initialize_controller(self, address):
        # check if controller is already initialized
        if address in self.i2c_devices:
            return self.i2c_devices[address]

        i2c_device = await self.platform.configure_i2c(address)
        self.i2c_devices[address] = i2c_device

        # initialise PCA9685/PCA9635
        i2c_device.i2c_write8(0x00, 0x11)  # set sleep
        i2c_device.i2c_write8(0x01, 0x04)  # configure output
        i2c_device.i2c_write8(0xFE, 130)   # set approx 50Hz
        await asyncio.sleep(.01)     # needed according to datasheet to sync PLL
        i2c_device.i2c_write8(0x00, 0x01)  # no more sleep
        await asyncio.sleep(.01)     # needed to end sleep according to datasheet
        return i2c_device

    async def configure_servo(self, number: str):
        """Configure servo."""
        try:
            i2c_address, servo_number = number.rsplit("-", 1)
        except ValueError:
            servo_number = number
            i2c_address = str(0x40)
        try:
            number_int = int(servo_number)
        except ValueError:
            raise ConfigFileError("Invalid servo number {} in {}.".format(servo_number, number),
                                  2, self.log.name if self.log else "")

        i2c_device = await self._initialize_controller(i2c_address)

        # check bounds
        if number_int < 0 or number_int > 15:
            raise ConfigFileError("Invalid number {} in {}. The controller only supports servos 0 to 15.".format(
                number_int, number), 1, self.log.name if self.log else "")

        return I2cServo(number_int, self.config, i2c_device)

    def stop(self):
        """Stop platform."""


class I2cServo(ServoPlatformInterface):

    """A servo hw device."""

    def __init__(self, number, config, i2c_device):
        """Initialise I2C hw servo."""
        self.log = logging.getLogger('I2cServo')
        self.number = number
        self.config = config
        self.i2c_device = i2c_device

    def go_to_position(self, position):
        """Move servo to position.

        Args:
        ----
            position: Position to set the servo. 0 to 1

        """
        # check bounds
        if position < 0 or position > 1:
            raise AssertionError("Position has to be between 0 and 1")

        # actual values depend on the controller. usually 150 to 600.
        # interpolate
        servo_min = self.config['servo_min']
        servo_max = self.config['servo_max']
        value = int(servo_min + position * (servo_max - servo_min))

        # set servo via i2c
        self.i2c_device.i2c_write8(0x06 + self.number * 4, 0)
        self.i2c_device.i2c_write8(0x07 + self.number * 4, 0)
        self.i2c_device.i2c_write8(0x08 + self.number * 4, value & 0xFF)
        self.i2c_device.i2c_write8(0x09 + self.number * 4, value >> 8)

    def set_speed_limit(self, speed_limit):
        """Not implemented."""

    def set_acceleration_limit(self, acceleration_limit):
        """Not implemented."""



"""Test score reels."""
from unittest.mock import MagicMock

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase


class TestScoreReels(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/score_reels/'

    def _synchronise_to_reel(self):
        pass

    def testOvershooting(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        chime1 = self.machine.coils["chime1"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        chime1.pulse = MagicMock(return_value=10)

        self._synchronise_to_reel()
        # reel desyncs while idle
        self.release_switch_and_run("score_1p_10_0", 0)

        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(0, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # resynced
        self.hit_switch_and_run("score_1p_10_0", 0)
        self.advance_time_and_run(10)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(0, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # advance one position
        self.machine.score_reels["score_1p_100"].set_destination_value(1)
        self.advance_time_and_run(.02)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # success
        self.release_switch_and_run("score_1p_100_0", 0)
        self.advance_time_and_run(10)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # the reel jumps back (for some reason)
        self.hit_switch_and_run("score_1p_100_0", 0)
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # resync
        self.release_switch_and_run("score_1p_100_0", 0)
        self.advance_time_and_run(10)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

    def testScoring(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        chime1 = self.machine.coils["chime1"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        chime1.pulse = MagicMock(return_value=10)

        self.start_game()
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)
        self.assertBallNumber(1)

        self.advance_time_and_run()
        self._synchronise_to_reel()
        self.machine.game.player.score = 110
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.hit_switch_and_run("score_1p_10k_0", 0)
        self.hit_switch_and_run("score_1p_1k_0", 0)
        self.release_switch_and_run("score_1p_100_0", 0)
        self.release_switch_and_run("score_1p_10_0", 0)

        self.advance_time_and_run(10)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(0, chime1.pulse.call_count)
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)

        self._synchronise_to_reel()
        self.machine.game.player.score = 11207
        self.advance_time_and_run(.005)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(0, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.advance_time_and_run(.02)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.advance_time_and_run(.02)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.advance_time_and_run(.02)
        self.assertEqual(1, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        self.release_switch_and_run("score_1p_10k_0", 0)
        self.release_switch_and_run("score_1p_1k_0", 0)

        self.advance_time_and_run(.17)
        self.assertEqual(2, player1_10.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(1, chime1.pulse.call_count)
        self.assertEqual(3, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(4, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(5, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(6, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(7, player1_10.pulse.call_count)

        self.advance_time_and_run(.2)
        self.assertEqual(1, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(8, player1_10.pulse.call_count)

        # it was stuck somewhere before 9
        self.advance_time_and_run(.2)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(9, player1_10.pulse.call_count)

        self.advance_time_and_run(.1)
        self.hit_switch_and_run("score_1p_10_9", 0)
        self.advance_time_and_run(.1)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)

        self.advance_time_and_run(.1)
        self.release_switch_and_run("score_1p_10_9", 0)
        self.hit_switch_and_run("score_1p_10_0", 0)
        self.advance_time_and_run(.1)

        # no more changes
        self.advance_time_and_run(10)
        self.assertEqual(1, player1_10k.pulse.call_count)
        self.assertEqual(1, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)
        self.assertEqual(1, chime1.pulse.call_count)

    def testAdvanceingFailure(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        self.start_game()

        self._synchronise_to_reel()
        self.machine.game.player.score += 110
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.release_switch_and_run("score_1p_10_0", 0)
        # switch for pos 0 stays on

        # it retries
        self.advance_time_and_run(.25)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # and again
        self.advance_time_and_run(.25)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(3, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        # finally
        self.release_switch_and_run("score_1p_100_0", 0)

        # no more retries
        self.advance_time_and_run(10)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(3, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

    def testThreePlayers(self):
        player1_10k = self.machine.coils["player1_10k"].hw_driver
        player1_1k = self.machine.coils["player1_1k"].hw_driver
        player1_100 = self.machine.coils["player1_100"].hw_driver
        player1_10 = self.machine.coils["player1_10"].hw_driver
        player2_10 = self.machine.coils["player2_10"].hw_driver
        player1_10k.pulse = MagicMock(return_value=10)
        player1_1k.pulse = MagicMock(return_value=10)
        player1_100.pulse = MagicMock(return_value=10)
        player1_10.pulse = MagicMock(return_value=10)
        player2_10.pulse = MagicMock(return_value=10)
        self.start_game()

        # add two more players
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertEqual(3, self.machine.game.num_players)

        self._synchronise_to_reel()
        self.machine.game.player.score += 110
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.release_switch_and_run("score_1p_10_0", 0)
        self.release_switch_and_run("score_1p_100_0", 0)

        # no more fires
        self.advance_time_and_run(10)
        self.machine_run()
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)

        # drain ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(.1)
        self.assertEqual(2, self.machine.game.player.number)

        self.machine.game.player.score += 20
        self.advance_time_and_run(.1)
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(1, player2_10.pulse.call_count)
        self.release_switch_and_run("score_2p_10_0", 0)

        self.advance_time_and_run(.3)
        self.machine_run()
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        # no more changes
        self.advance_time_and_run(10)
        self.machine_run()
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(1, player1_100.pulse.call_count)
        self.assertEqual(1, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        # drain ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(.15)
        self.assertEqual(3, self.machine.game.player.number)

        # player3 reuses the reels from player 1. machine resets them
        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(2, player1_100.pulse.call_count)
        self.assertEqual(2, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        for i in range(7):
            self.advance_time_and_run(.2)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(9, player1_100.pulse.call_count)
        self.assertEqual(9, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        self.hit_switch_and_run("score_1p_10_9", 0)
        self.hit_switch_and_run("score_1p_100_9", 0)

        self.advance_time_and_run(.2)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(10, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)

        self.hit_switch_and_run("score_1p_10_0", 0)
        self.hit_switch_and_run("score_1p_100_0", 0)
        self.release_switch_and_run("score_1p_10_9", 0)
        self.release_switch_and_run("score_1p_100_9", 0)

        self.advance_time_and_run(10)

        self.assertEqual(0, player1_10k.pulse.call_count)
        self.assertEqual(0, player1_1k.pulse.call_count)
        self.assertEqual(10, player1_100.pulse.call_count)
        self.assertEqual(10, player1_10.pulse.call_count)
        self.assertEqual(2, player2_10.pulse.call_count)


class TestScoreReelsVirtual(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_platform(self):
        return "smart_virtual"

    def get_machine_path(self):
        return 'tests/machine_files/score_reels/'

    def testScoringVirtual(self):
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")

        # wait for reset of score reels
        self.advance_time_and_run(4)
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)

        self.advance_time_and_run()
        self.assertSwitchState("score_1p_10_0", 1)
        self.assertSwitchState("score_1p_100_0", 1)
        self.assertSwitchState("score_1p_10_9", 0)
        self.assertSwitchState("score_1p_100_9", 0)

        self.machine.game.player.score += 110
        self.advance_time_and_run(10)

        self.assertSwitchState("score_1p_10_0", 0)
        self.assertSwitchState("score_1p_100_0", 0)
        self.assertSwitchState("score_1p_10_9", 0)
        self.assertSwitchState("score_1p_100_9", 0)

        self.machine.game.player.score = 990
        self.advance_time_and_run(10)

        self.assertSwitchState("score_1p_10_0", 0)
        self.assertSwitchState("score_1p_100_0", 0)
        self.assertSwitchState("score_1p_10_9", 1)
        self.assertSwitchState("score_1p_100_9", 1)

        self.machine.game.player.score = 1900
        self.advance_time_and_run(10)

        self.assertSwitchState("score_1p_10_0", 1)
        self.assertSwitchState("score_1p_100_0", 0)
        self.assertSwitchState("score_1p_10_9", 0)
        self.assertSwitchState("score_1p_100_9", 1)

    def test_inactive_player_scoring(self):
        self.start_two_player_game()
        self.assertGameIsRunning()
        self.assertPlayerNumber(1)

        self.advance_time_and_run()
        self.assertEqual([None, 0, 0, 0, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 0], self.machine.score_reel_groups["player2"].desired_value_list)

        # increasing score for player two should only affect score of that player
        self.machine.game.player_list[1].score = 3333
        self.advance_time_and_run()
        self.assertEqual([None, 0, 0, 0, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 3], self.machine.score_reel_groups["player2"].desired_value_list)

        self.machine.game.player_list[0].score = 2222
        self.advance_time_and_run()
        self.assertEqual([None, 2, 2, 2, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 3], self.machine.score_reel_groups["player2"].desired_value_list)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(1)

        self.machine.game.player_list[0].score = 4444
        self.advance_time_and_run()
        self.assertEqual([None, 4, 4, 4, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 3], self.machine.score_reel_groups["player2"].desired_value_list)

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)

        self.drain_all_balls()
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)

        self.drain_all_balls()
        self.assertPlayerNumber(2)
        self.assertBallNumber(3)

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        # start game. we cannot use start_two_player_game because score reels block the start for some seconds
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsRunning()
        self.assertEqual(0, self.machine.game.num_players)
        self.advance_time_and_run()
        self.assertEqual(1, self.machine.game.num_players)
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.game.num_players)
        self.assertPlayerNumber(1)

        self.assertEqual([None, 0, 0, 0, 0], self.machine.score_reel_groups["player1"].desired_value_list)
        self.assertEqual([None, 0], self.machine.score_reel_groups["player2"].desired_value_list)



"""Contains the Credit (coin play) mode code."""

from math import floor

from mpf.core.placeholder_manager import NativeTypeTemplate
from mpf.core.settings_controller import SettingEntry

from mpf.core.mode import Mode


class Credits(Mode):

    """Mode which manages the credits and prevents the game from starting without credits."""

    __slots__ = ["data_manager", "earnings", "credit_units_per_game", "credit_unit", "pricing_tiers",
                 "credit_units_for_pricing_tiers", "reset_pricing_tier_count_this_game", "credits_config"]

    def __init__(self, *args, **kwargs):
        """Initialise credits mode."""
        self.data_manager = None
        self.earnings = None

        self.credit_units_per_game = None
        self.credit_unit = None
        self.pricing_tiers = None
        self.credit_units_for_pricing_tiers = None
        self.reset_pricing_tier_count_this_game = None
        self.credits_config = None
        super().__init__(*args, **kwargs)

    def mode_init(self):
        """Initialise mode."""
        self.data_manager = self.machine.create_data_manager('earnings')
        self.earnings = self.data_manager.get_data()

        # prevent crashes because of data corruption
        if not isinstance(self.earnings, dict):
            self.earnings = {}

        self.credit_units_per_game = 0
        self.credit_unit = 0
        self.pricing_tiers = set()
        self.credit_units_for_pricing_tiers = 0
        self.reset_pricing_tier_count_this_game = False

        self.credits_config = self.machine.config_validator.validate_config(
            config_spec='credits',
            source=self.machine.config.get('credits', {}),
            section_name='credits')

        for event in self.credits_config['reset_earnings_events']:
            self.machine.events.add_handler(event, self._reset_earnings)

        for event in self.credits_config['reset_credits_events']:
            self.machine.events.add_handler(event, self._reset_credits)

        # add setting
        self.machine.settings.add_setting(SettingEntry("free_play", "Free Play", 500,
                                                       "free_play", self.credits_config['free_play'],
                                                       {True: "Free Play", False: "Credit Play"}))

    def mode_start(self, **kwargs):
        """Start mode."""
        self.add_mode_event_handler('enable_free_play',
                                    self.enable_free_play)
        self.add_mode_event_handler('enable_credit_play',
                                    self.enable_credit_play)
        self.add_mode_event_handler('toggle_credit_play',
                                    self.toggle_credit_play)
        self.add_mode_event_handler('slam_tilt',
                                    self.clear_all_credits)

        if self.machine.settings.get_setting_value("free_play"):
            self.enable_free_play(post_event=False)
        else:
            self._calculate_credit_units()
            self._calculate_pricing_tiers()
            self.enable_credit_play(post_event=False)

    def _reset_earnings(self, **kwargs):
        """Reset earnings."""
        del kwargs
        self.log.info("Resetting all earnings.")
        self.earnings = {}
        self.data_manager.save_all(data=self.earnings)

    def _reset_credits(self, **kwargs):
        """Reset credits."""
        del kwargs
        self.log.info("Resetting all credits.")
        self.clear_all_credits()

    def mode_stop(self, **kwargs):
        """Stop mode."""
        self._set_free_play_string()
        self._disable_credit_handlers()

    def _calculate_credit_units(self):
        # "credit units" are how we handle fractional credits (since most
        # pinball machines show credits as fractions instead of decimals).
        # We convert everything to the smallest coin unit and then track
        # how many of those a game takes. So price of $0.75 per game with a
        # quarter slot means a credit unit is 0.25 and the game needs 3 credit
        # units to start. This is all hidden from the player

        # We need to calculate it differently depending on how the coin switch
        # values relate to game cost.

        if self.credits_config['switches']:
            min_currency_value = min(x['value'].evaluate([]) for x in
                                     self.credits_config['switches'])
        else:
            try:
                min_currency_value = (
                    self.credits_config['pricing_tiers'][0]['price'].evaluate([]))
            except IndexError:
                min_currency_value = 1

        try:
            price_per_game = self.credits_config['pricing_tiers'][0]['price'].evaluate([])
            if self.credits_config['pricing_tiers'][0]['credits'] != NativeTypeTemplate(1, self.machine):
                raise AssertionError("First pricing_tier entry has to give exactly one credits.")
        except IndexError:
            price_per_game = 1

        if min_currency_value == price_per_game:
            self.credit_unit = min_currency_value

        elif min_currency_value < price_per_game:
            self.credit_unit = price_per_game - min_currency_value
            if self.credit_unit > min_currency_value:
                self.credit_unit = min_currency_value

        elif min_currency_value > price_per_game:
            self.credit_unit = min_currency_value - price_per_game
            if self.credit_unit > price_per_game:
                self.credit_unit = price_per_game

        self.debug_log("Calculated the credit unit to be %s based on a minimum"
                       "currency value of %s and a price per game of %s",
                       self.credit_unit, min_currency_value, price_per_game)

        self.credit_units_per_game = int(price_per_game / self.credit_unit)

        self.info_log("Credit units per game: %s", self.credit_units_per_game)

    def _calculate_pricing_tiers(self):
        # pricing tiers are calculated with a set of tuples which indicate the
        # credit units for the price break as well as the "bump" in credit
        # units that should be added once that break is passed.

        index = 0
        for pricing_tier in self.credits_config['pricing_tiers']:
            price = pricing_tier['price'].evaluate([])
            credit_units = price / self.credit_unit
            credits_in_tier = pricing_tier['credits'].evaluate([])
            actual_credit_units = self.credit_units_per_game * credits_in_tier
            bonus = actual_credit_units - credit_units

            self.machine.variables.set_machine_var("price_per_game_raw_{}".format(index), price)
            self.machine.variables.set_machine_var("price_per_game_string_{}".format(index),
                                                   self.credits_config['price_tier_template'].format(
                                                       price=price, credits=credits_in_tier))

            self.debug_log("Pricing Tier Bonus. Price: %s, Credits: %s. "
                           "Credit units for this tier: %s, Credit units this "
                           "tier buys: %s, Bonus bump needed: %s",
                           price, credits_in_tier,
                           credit_units, actual_credit_units, bonus)

            self.pricing_tiers.add((credit_units, bonus))
            index += 1

    def enable_credit_play(self, post_event=True, **kwargs):
        """Enable credits play."""
        del kwargs

        credit_units = self._get_credit_units()

        if self.credits_config['persist_credits_while_off_time']:
            self.machine.variables.configure_machine_var(name='credit_units', persist=True,
                                                         expire_secs=self.credits_config[
                                                             'persist_credits_while_off_time'])
        self.machine.variables.set_machine_var(name='credit_units', value=credit_units)

        '''machine_var: credit_units

        desc: How many credit units are on the machine. Note that credit units
        are not useful for display purposes since they represent the number of
        credits in a ration related to the lowest common denominator of the
        partial credit fraction. See the related *credits_string* and
        *credits_value* machine variables for more useful formats.
        '''

        self.machine.variables.set_machine_var('credits_string', ' ')
        # doc string is in machine.py for credits_string

        self.machine.settings.set_setting_value('free_play', False)

        self.machine.variables.set_machine_var('credits_value', '0')
        '''machine_var: credits_value

        desc: The human readable string form which shows the number value of
        how many credits are on the machine, including whole and fractional
        credits, for example "1" or "2 1/2" or "3 3/4".

        If you want the full string with the word "CREDITS" in it, use the
        "credits_string" machine variable.
        '''

        self.machine.variables.set_machine_var('credits_whole_num', 0)
        '''machine_var: credits_whole_num

        desc: The whole number portion of the total credits on the machine.
        For example, if the machine has 3 1/2 credits, this value is "3".
        '''

        self.machine.variables.set_machine_var('credits_numerator', 0)
        '''machine_var: credits_numerator

        desc: The numerator portion of the total credits on the machine.
        For example, if the machine has 4 1/2 credits, this value is "1".
        '''
        self.machine.variables.set_machine_var('credits_denominator', 0)
        '''machine_var: credits_whole_num

        desc: The denominator portion of the total credits on the machine.
        For example, if the machine has 4 1/2 credits, this value is "2".
        '''

        self._update_credit_strings()

        self._enable_credit_handlers()

        # prevent duplicate handlers
        self._remove_event_handlers()

        # setup event handlers
        self.add_mode_event_handler('player_add_request',
                                    self._player_add_request)
        self.add_mode_event_handler('request_to_start_game',
                                    self._request_to_start_game)
        self.add_mode_event_handler('player_added',
                                    self._player_added)
        self.add_mode_event_handler('mode_game_started',
                                    self._game_started)
        self.add_mode_event_handler('mode_game_stopped',
                                    self._game_ended)
        self.add_mode_event_handler('ball_starting',
                                    self._ball_starting)
        if post_event:
            self.machine.events.post('enabling_credit_play')
        '''event: enabling_credit_play
        desc: The game is no longer on free play. Credits are required to
        start a game. This event is also posted on MPF boot if the credits mode
         is enabled and the game is not set to free play.
        '''

    def _remove_event_handlers(self):
        """Remove event handlers."""
        self.machine.events.remove_handler(self._player_add_request)
        self.machine.events.remove_handler(self._request_to_start_game)
        self.machine.events.remove_handler(self._player_added)
        self.machine.events.remove_handler(self._game_ended)
        self.machine.events.remove_handler(self._game_started)
        self.machine.events.remove_handler(self._ball_starting)

    def enable_free_play(self, post_event=True, **kwargs):
        """Enable free play."""
        del kwargs
        self.machine.settings.set_setting_value('free_play', True)

        for index in range(len(self.credits_config['pricing_tiers'])):
            self.machine.variables.remove_machine_var("price_per_game_raw_{}".format(index))
            self.machine.variables.remove_machine_var("price_per_game_string_{}".format(index))

        self._remove_event_handlers()

        self._disable_credit_handlers()

        self._update_credit_strings()

        if post_event:
            self.machine.events.post('enabling_free_play')
        '''event: enabling_free_play
        desc: Credits are no longer required to start a game. This event is
        also posted on MPF boot if the credits mode is enabled and the game is
        set to free play.
        '''

    def toggle_credit_play(self, **kwargs):
        """Toggle between free and credits play."""
        del kwargs

        if self.machine.settings.get_setting_value('free_play'):
            self.enable_credit_play()
        else:
            self.enable_free_play()

    def _get_credit_units(self):
        credit_units = self.machine.variables.get_machine_var('credit_units')
        if not credit_units:
            credit_units = 0
        return credit_units

    def _player_add_request(self, **kwargs):
        del kwargs
        if (self._get_credit_units() >=
                self.credit_units_per_game):
            self.info_log("Received request to add player. Request Approved. Sufficient credits available.")
            return True

        self.info_log("Received request to add player. Request Denied. Not enough credits available.")
        self.machine.events.post("not_enough_credits")
        '''event: not_enough_credits
        desc: A player has pushed the start button, but the game is not set
        to free play and there are not enough credits to start a game or
        add a player.
        '''
        return False

    def _request_to_start_game(self, **kwargs):
        del kwargs
        if (self._get_credit_units() >=
                self.credit_units_per_game):
            self.info_log("Received request to start game. Request Approved. Sufficient credits available.")
            return True

        self.info_log("Received request to start game. Request Denied. Not enough credits available.")
        self.machine.events.post("not_enough_credits")
        # event docstring covered in _player_add_request() method
        return False

    def _player_added(self, **kwargs):
        del kwargs
        new_credit_units = (self._get_credit_units() -
                            self.credit_units_per_game)

        if new_credit_units < 0:
            self.warning_log("Somehow credit units went below 0?!? Resetting "
                             "to 0.")
            new_credit_units = 0

        self.machine.variables.set_machine_var('credit_units', new_credit_units)
        self._update_credit_strings()

    def _enable_credit_handlers(self):
        for switch_settings in self.credits_config['switches']:
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch_settings['switch'],
                callback=self._credit_switch_callback,
                callback_kwargs={'value': switch_settings['value'].evaluate([]),
                                 'audit_class': switch_settings['type']})

        for switch in self.credits_config['service_credits_switch']:
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch,
                callback=self._service_credit_callback)

        for event_settings in self.credits_config['events']:
            self.machine.events.add_handler(
                event=event_settings['event'],
                handler=self._credit_event_callback,
                credits=event_settings['credits'],
                audit_class=event_settings['type'])

    def _disable_credit_handlers(self):
        for switch_settings in self.credits_config['switches']:
            self.machine.switch_controller.remove_switch_handler(
                switch_name=switch_settings['switch'].name,
                callback=self._credit_switch_callback)

        self.machine.events.remove_handler(self._credit_event_callback)

        for switch in self.credits_config['service_credits_switch']:
            self.machine.switch_controller.remove_switch_handler(
                switch_name=switch.name,
                callback=self._service_credit_callback)

    def _credit_switch_callback(self, value, audit_class):
        self.info_log("Credit switch hit. Credit Added. Value: %s. Type: %s", value, audit_class)
        self._add_credit_units(credit_units=value / self.credit_unit)
        self._audit(value, audit_class)
        self._reset_timeouts()

    # credits is a built-in in python
    # pylint: disable-msg=redefined-builtin
    def _credit_event_callback(self, credits, audit_class, **kwargs):
        credits_value = credits.evaluate(kwargs)
        self.info_log("Credit event hit. Credit Added. Credits: %s. Type: %s", credits_value, audit_class)
        self._add_credit_units(credit_units=credits_value * self.credit_units_per_game, price_tiering=False)
        self._audit(credits_value, audit_class)
        self._reset_timeouts()

    def _service_credit_callback(self):
        self.info_log("Service Credit Added. Value: 1.")
        self.add_credit(price_tiering=False)
        self._audit(1, 'service_credit')

    def _add_credit_units(self, credit_units, price_tiering=True):
        self.debug_log("Adding %s credit_units. Price tiering: %s",
                       credit_units, price_tiering)

        previous_credit_units = self._get_credit_units()
        total_credit_units = credit_units + previous_credit_units

        # check for pricing tier
        if price_tiering:
            self.credit_units_for_pricing_tiers += credit_units
            bonus_credit_units = 0
            for tier_credit_units, bonus in self.pricing_tiers:
                if self.credit_units_for_pricing_tiers % tier_credit_units == 0:
                    bonus_credit_units += bonus

            total_credit_units += bonus_credit_units

        max_credit_units = (self.credits_config['max_credits'].evaluate([]) *
                            self.credit_units_per_game)

        if max_credit_units and total_credit_units > max_credit_units:
            self.info_log("Max credits reached.")
            self._update_credit_strings()
            self.machine.events.post('max_credits_reached')
            '''event: max_credits_reached
            desc: Credits have just been added to the machine, but the
            configured maximum number of credits has been reached.'''
            self.machine.variables.set_machine_var('credit_units', max_credit_units)

        if max_credit_units <= 0 or max_credit_units > previous_credit_units:
            self.info_log("Credit units added")
            self.machine.variables.set_machine_var('credit_units', total_credit_units)
            self._update_credit_strings()
            self.machine.events.post('credits_added')
            '''event: credits_added
            desc: Credits (or partial credits) have just been added to the
            machine.'''

    def add_credit(self, price_tiering=True):
        """Add a single credit to the machine.

        Args:
        ----
            price_tiering: Boolean which controls whether this credit will be
                eligible for the pricing tier bonuses. Default is True.

        """
        self._add_credit_units(self.credit_units_per_game, price_tiering)

    def _reset_pricing_tier_credits(self):
        if not self.reset_pricing_tier_count_this_game:
            self.info_log("Resetting pricing tier credit count.")
            self.credit_units_for_pricing_tiers = 0
            self.reset_pricing_tier_count_this_game = True

    def _ball_starting(self, player, ball, **kwargs):
        del kwargs
        if player == 1 and ball == 2:
            self._reset_pricing_tier_credits()

    def _set_free_play_string(self):
        display_string = self.credits_config['free_play_string']
        self.machine.variables.set_machine_var('credits_string', display_string)

    def _update_credit_strings(self):
        if self.machine.settings.get_setting_value('free_play'):
            self._set_free_play_string()
            return

        machine_credit_units = self._get_credit_units()
        if self.credit_units_per_game > 0:
            whole_num = int(floor(machine_credit_units /
                            self.credit_units_per_game))
            numerator = int(machine_credit_units % self.credit_units_per_game)
            denominator = int(self.credit_units_per_game)
        else:
            whole_num = 0
            numerator = 0
            denominator = 0

        if numerator:
            if whole_num:
                display_fraction = '{} {}/{}'.format(whole_num, numerator,
                                                     denominator)
            else:
                display_fraction = '{}/{}'.format(numerator, denominator)

        else:
            display_fraction = str(whole_num)

        display_string = '{} {}'.format(self.credits_config['credits_string'], display_fraction)
        self.machine.variables.set_machine_var('credits_string', display_string)
        self.machine.variables.set_machine_var('credits_value', display_fraction)
        self.machine.variables.set_machine_var('credits_whole_num', whole_num)
        self.machine.variables.set_machine_var('credits_numerator', numerator)
        self.machine.variables.set_machine_var('credits_denominator', denominator)

    def _audit(self, value, audit_class):
        if audit_class not in self.earnings:
            self.earnings[audit_class] = dict()
            self.earnings[audit_class]['total_value'] = 0
            self.earnings[audit_class]['count'] = 0

        self.earnings[audit_class]['total_value'] += value
        self.earnings[audit_class]['count'] += 1

        self.data_manager.save_all(data=self.earnings)

    def _game_started(self, **kwargs):
        del kwargs
        self.debug_log("Removing credit clearing delays.")
        self.delay.remove('clear_fractional_credits')
        self.delay.remove('clear_all_credits')

    def _reset_timeouts(self):
        if self.credits_config['fractional_credit_expiration_time']:
            self.debug_log("Adding delay to clear fractional credits")
            self.delay.reset(
                ms=self.credits_config['fractional_credit_expiration_time'],
                callback=self._clear_fractional_credits,
                name='clear_fractional_credits')

        if self.credits_config['credit_expiration_time']:
            self.debug_log("Adding delay to clear credits")
            self.delay.reset(
                ms=self.credits_config['credit_expiration_time'],
                callback=self.clear_all_credits,
                name='clear_all_credits')

    def _game_ended(self, **kwargs):
        del kwargs
        self._reset_timeouts()

        self.reset_pricing_tier_count_this_game = False

    def _clear_fractional_credits(self):
        self.info_log("Clearing fractional credits.")

        credit_units = self._get_credit_units()
        credit_units -= credit_units % self.credit_units_per_game

        self.machine.variables.set_machine_var('credit_units', credit_units)
        self._update_credit_strings()

    def clear_all_credits(self, **kwargs):
        """Clear all credits."""
        del kwargs
        self.info_log("Clearing all credits.")
        self.machine.variables.set_machine_var('credit_units', 0)
        self._update_credit_strings()



"""Count balls using an entrance switch."""
import asyncio

from mpf.core.delays import DelayManager
from mpf.devices.ball_device.physical_ball_counter import PhysicalBallCounter, BallEntranceActivity, \
    BallLostActivity


class EntranceSwitchCounter(PhysicalBallCounter):

    """Count balls using an entrance switch."""

    __slots__ = ["recycle_secs", "recycle_clear_time", "_settle_delay"]

    def __init__(self, ball_device, config):
        """Initialise entrance switch counter."""
        for option in ["entrance_switch", "entrance_switch_ignore_window_ms", "entrance_switch_full_timeout",
                       "ball_capacity"]:
            if option not in config and option in ball_device.config:
                config[option] = ball_device.config[option]
        super().__init__(ball_device, config)
        self._settle_delay = DelayManager(self.machine)

        self.config = self.machine.config_validator.validate_config("ball_device_counter_entrance_switches",
                                                                    self.config)

        self.recycle_secs = self.config['entrance_switch_ignore_window_ms'] / 1000.0
        self.recycle_clear_time = {}

        for switch in self.config['entrance_switch']:
            # Configure switch handlers for entrance switch activity
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=1,
                ms=0,
                callback=self._entrance_switch_handler,
                callback_kwargs={"switch_name": switch.name})

            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, state=0,
                ms=0,
                callback=self._entrance_switch_released_handler,
                callback_kwargs={"switch_name": switch.name})

        if self.config['entrance_switch_full_timeout'] and self.config['ball_capacity']:
            if len(self.config['entrance_switch']) > 1:
                raise AssertionError("entrance_switch_full_timeout not supported with multiple entrance switches.")
            self.machine.switch_controller.add_switch_handler_obj(
                switch=self.config['entrance_switch'][0], state=1,
                ms=self.config['entrance_switch_full_timeout'],
                callback=self._entrance_switch_full_handler)

        # Handle initial ball count with entrance_switch. If there is a ball on the entrance_switch at boot
        # assume that we are at max capacity.
        if (self.config['ball_capacity'] and self.config['entrance_switch_full_timeout'] and
                self.machine.switch_controller.is_active(self.config['entrance_switch'][0],
                                                         ms=self.config['entrance_switch_full_timeout'])):
            self._last_count = self.config['ball_capacity']
        else:
            self._last_count = 0
        self._count_stable.set()

        # TODO validate that we are not used with mechanical eject
        if not self.config['ball_capacity']:
            self.ball_device.raise_config_error("Need ball capacity if there are no switches.", 2)
        elif self.ball_device.config.get('ball_switches'):
            self.ball_device.raise_config_error("Cannot use capacity and ball switches.", 3)

    @property
    def capacity(self):
        """Return capacity under normal circumstances (i.e. without jam switches)."""
        return self.config['ball_capacity']

    def is_jammed(self) -> bool:
        """Return False because this device can not know if it is jammed."""
        return False

    def is_count_unreliable(self) -> bool:
        """Return False because this device can not know if it is jammed."""
        return False

    def received_entrance_event(self):
        """Handle entrance event."""
        self._entrance_switch_handler("event")

    def _recycle_passed(self, switch):
        self.recycle_clear_time[switch] = None

    def _entrance_switch_handler(self, switch_name):
        """Add a ball to the device since the entrance switch has been hit."""
        # always invalidate count if this has been triggered by a real switch
        if switch_name != "event":
            self.invalidate_count()
        # If recycle is ongoing, do nothing
        if self.recycle_clear_time.get(switch_name, False):
            self.debug_log("Entrance switch hit within ignore window, taking no action")
            return
        # If a recycle time is configured, set a timeout to prevent future entrance activity
        if self.recycle_secs:
            self.recycle_clear_time[switch_name] = self.machine.clock.get_time() + self.recycle_secs
            self.machine.clock.loop.call_at(self.recycle_clear_time[switch_name], self._recycle_passed, switch_name)

        self.debug_log("Entrance switch hit")
        if self.config['ball_capacity'] and self.config['ball_capacity'] <= self._last_count:
            # do not count beyond capacity
            self.ball_device.log.warning("Device received balls but is already full!")
        elif self.config['ball_capacity'] and self.config['entrance_switch_full_timeout'] and \
                self.config['ball_capacity'] == self._last_count + 1:
            # wait for entrance_switch_full_timeout before setting the device to full capacity
            self._settle_delay.remove("count_stable")
        else:
            # increase count
            self._settle_delay.reset(self.config['settle_time_ms'], self.mark_count_as_stable_and_trigger_activity,
                                     "count_stable")
            self._last_count += 1
            self.record_activity(BallEntranceActivity())

    def _entrance_switch_released_handler(self, switch_name):
        """Entrance switch has been released."""
        del switch_name
        # count is stable once switch is inactive
        # (or if it stays on the switch long enough; see _entrance_switch_full_handler)
        self._count_stable.set()

    def _entrance_switch_full_handler(self):
        # a ball is sitting on the entrance_switch. assume the device is full
        self._count_stable.set()
        new_balls = self.config['ball_capacity'] - self._last_count
        self.mark_count_as_stable_and_trigger_activity()
        if new_balls > 0:
            self.debug_log("Ball is sitting on entrance_switch. Assuming "
                           "device is full. Adding %s balls and setting balls"
                           "to %s", new_balls, self.config['ball_capacity'])
            self._last_count += new_balls
            for _ in range(new_balls):
                self.record_activity(BallEntranceActivity())

    def count_balls_sync(self) -> int:
        """Return the number of balls entered."""
        assert self._last_count is not None
        if self.config['ball_capacity'] and self.config['ball_capacity'] == self._last_count:
            # we are at capacity. this is fine
            pass
        elif self.config['ball_capacity'] and self.config['entrance_switch_full_timeout'] and \
            self.config['ball_capacity'] == self._last_count + 1 and \
            self.machine.switch_controller.is_active(self.config['entrance_switch'][0],
                                                     ms=self.config['entrance_switch_full_timeout']):
            # can count when entrance switch is active for at least entrance_switch_full_timeout
            pass
        elif not self.is_ready_to_receive:
            # cannot count when the entrance_switch is still active
            raise ValueError

        return self._last_count

    async def wait_for_ball_to_leave(self):
        """Wait for a ball to leave."""
        await self.wait_for_count_stable()
        # wait 10ms
        done_future = asyncio.ensure_future(asyncio.sleep(0.01))
        done_future.add_done_callback(self._ball_left)
        return done_future

    def _ball_left(self, future):
        del future
        self._last_count -= 1
        self.record_activity(BallLostActivity())
        self.trigger_activity()

    @property
    def is_ready_to_receive(self):
        """Return true if entrance switch is inactive."""
        return not all(self.machine.switch_controller.is_active(switch) for switch in self.config['entrance_switch'])

    async def wait_for_ready_to_receive(self):
        """Wait until all entrance switch are inactive."""
        while True:
            if self.is_ready_to_receive:
                return True

            await self.machine.switch_controller.wait_for_any_switch(
                switches=self.config['entrance_switch'],
                state=0, only_on_change=True)



from mpf.tests.MpfTestCase import MpfTestCase, MagicMock, patch


class TestRpiDmd(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/rpi_dmd/'

    def get_platform(self):
        # no force platform.
        return False

    def setUp(self):
        modules = {
            'PIL': MagicMock(),
            'PIL.Image': MagicMock(),
        }
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        Image = patch('mpf.platforms.rpi_dmd.Image')
        self.image = Image.start()

        RGBMatrix = patch('mpf.platforms.rpi_dmd.RGBMatrix')
        RGBMatrixOptions = patch('mpf.platforms.rpi_dmd.RGBMatrixOptions')
        self.rgbmatrixoptions = RGBMatrixOptions.start()
        self.rgbmatrix = RGBMatrix.start()
        self.rgbmatrix_instance = MagicMock()
        self.rgbmatrix.return_value = self.rgbmatrix_instance
        self.addCleanup(self.rgbmatrix.stop)
        super().setUp()

    def tearDown(self):
        self.module_patcher.stop()
        super().tearDown()

    def test_rpi_dmd(self):
        data = bytes([0x02] * 32 * 32 * 3)
        self.machine.rgb_dmds["rpi_dmd"].update(data)
        self.advance_time_and_run(.1)

        self.assertTrue(len(self.rgbmatrix_instance.SetImage.mock_calls) > 0)
        last_call = self.rgbmatrix_instance.SetImage.mock_calls[1]
        frombytes_call = last_call[1][0]
        frombytes_arg = frombytes_call.mock_calls[0][1][0]
        self.assertEqual(data, frombytes_arg)



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallHold(MpfTestCase):

    def get_config_file(self):
        return 'test_ball_holds.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_holds/'

    def _missing_ball(self, **kwargs):
        del kwargs
        self._missing += 1

    def _ball_enter(self, new_balls, unclaimed_balls, **kwargs):
        del unclaimed_balls
        del kwargs
        self._enter += new_balls

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def _collecting_balls_complete_handler(self, **kwargs):
        del kwargs
        self._collecting_balls_complete = 1

    def test_ball_hold_in_mode(self):
        # start mode
        self.post_event("start_mode1")

        # mode loaded. ball_hold2 should be enabled
        self.assertTrue(self.machine.ball_holds["hold_test2"].enabled)

        # stop mode
        self.post_event("stop_mode1")

        # mode stopped. should ball_hold be disabled
        self.assertFalse(self.machine.ball_holds["hold_test2"].enabled)

        # start mode (again)
        self.post_event("start_mode1")

        # mode loaded. ball_hold2 should be enabled
        self.assertTrue(self.machine.ball_holds["hold_test2"].enabled)

        # stop mode
        self.post_event("stop_mode1")

        # mode stopped. should ball_hold be disabled
        self.assertFalse(self.machine.ball_holds["hold_test2"].enabled)

    def test_hold_and_release_at_game_end(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        trough = self.machine.ball_devices['test_trough']
        launcher = self.machine.ball_devices['test_launcher']
        hold = self.machine.ball_devices['test_hold']
        hold_logic = self.machine.ball_holds['hold_test']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield', self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing', self._missing_ball)

        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, trough.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # start a game
        self.machine.switch_controller.process_switch("s_start", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)

        # trough ejects
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        self.machine.switch_controller.process_switch("s_playfield_active", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield_active", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.mock_event("yes")
        self.mock_event("no")
        self.post_event("test_conditional_event")
        self.assertEventCalled("no")
        self.assertEventNotCalled("yes")

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()

        # ball enters hold
        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, hold.balls)

        self.mock_event("yes")
        self.mock_event("no")
        self.post_event("test_conditional_event")
        self.assertEventNotCalled("no")
        self.assertEventCalled("yes")

        # request another ball
        self.machine.playfield.add_ball(1)
        self.advance_time_and_run()

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(1, hold_logic.balls_held)
        self._captured = 0

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        self.machine.switch_controller.process_switch("s_playfield_active", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield_active", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        # ball drains
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self._captured = 0
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # hold should eject all balls
        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        self.assertTrue(coil3.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, hold.balls)
        self.assertEqual(0, hold_logic.balls_held)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        # game did not end because ball has not drained
        self.assertIsNotNone(self.machine.game)

        # ball also drains
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        # game ends
        self.assertIsNone(self.machine.game)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        self.advance_time_and_run(100)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

    def test_hold_full_and_release(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        trough = self.machine.ball_devices['test_trough']
        hold = self.machine.ball_devices['test_hold']
        hold_logic = self.machine.ball_holds['hold_test']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield', self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing', self._missing_ball)
        self.machine.events.add_handler('collecting_balls_complete', self._collecting_balls_complete_handler)

        hold_logic.enable()

        self._enter = 0
        self._captured = 0
        self._missing = 0
        self._collecting_balls_complete = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, trough.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertFalse(hold_logic.is_full())
        self.assertEqual(2, trough.available_balls)

        # hold captures a first random ball
        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 1)
        self.advance_time_and_run(1)
        assert not coil3.pulse.called
        self.assertFalse(hold_logic.is_full())

        # self.assertEqual(1, trough.available_balls)
        self.assertEqual(1, hold.available_balls)

        # hold captures a second random ball
        self.machine.switch_controller.process_switch("s_ball_switch_hold2", 1)
        self.advance_time_and_run(1)
        assert not coil3.pulse.called
        self.assertTrue(hold_logic.is_full())
        # self.assertEqual(0, trough.available_balls)
        self.assertEqual(2, hold.available_balls)

        # hold captures a third random ball
        self.machine.switch_controller.process_switch("s_ball_switch_hold3", 1)
        self.advance_time_and_run(1)

        # it should eject it right away
        self.assertTrue(coil3.pulse.called)
        coil3.pulse = MagicMock()
        self.assertTrue(hold_logic.is_full())
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_ball_switch_hold3", 0)
        self.advance_time_and_run(11)
        self.assertTrue(hold_logic.is_full())
        self.assertEqual(2, hold.available_balls)

        hold_logic.release_all()
        self.advance_time_and_run(1)
        self.assertEqual(0, hold.available_balls)
        self.assertTrue(coil3.pulse.called)
        coil3.pulse = MagicMock()
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_ball_switch_hold2", 0)
        self.advance_time_and_run(1)
        self.assertFalse(hold_logic.is_full())

        self.advance_time_and_run(11)
        self.assertTrue(coil3.pulse.called)
        coil3.pulse = MagicMock()
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 0)
        self.advance_time_and_run(11)
        assert not coil3.pulse.called

    def test_eject_to_hold(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        trough = self.machine.ball_devices['test_trough']
        launcher = self.machine.ball_devices['test_launcher']
        hold = self.machine.ball_devices['test_hold']
        hold_logic = self.machine.ball_holds['hold_test']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield', self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing', self._missing_ball)
        self.machine.events.add_handler('collecting_balls_complete', self._collecting_balls_complete_handler)

        self._enter = 0
        self._captured = 0
        self._missing = 0
        self._collecting_balls_complete = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, trough.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # start a game
        self.machine.switch_controller.process_switch("s_start", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)

        # trough ejects
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()

        # ball directly enters the hold
        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, hold.balls)

        # request another ball
        self.machine.playfield.add_ball(1)
        self.advance_time_and_run()

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(1, hold_logic.balls_held)
        self._captured = 0

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        self.machine.switch_controller.process_switch("s_playfield_active", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield_active", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(1, hold.available_balls)

        # request a release of one ball from hold via event
        self.machine.events.post("release_test")
        # since we are not using a multi ball increase the balls_in_play manually
        self.assertEqual(1, self.machine.game.balls_in_play)
        self.machine.game.balls_in_play += 1
        self.advance_time_and_run(1)

        # hold should eject a ball
        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        self.assertTrue(coil3.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, hold.balls)
        self.assertEqual(0, hold_logic.balls_held)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        # ball drains instantly. one left on pf
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.game.balls_in_play)

        # other ball hits some pf switches
        self.machine.switch_controller.process_switch("s_playfield_active", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield_active", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self._captured = 0
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        self.assertEqual(0, self._collecting_balls_complete)
        self.assertEqual(1, self.machine.game.balls_in_play)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()

        # we add another ball
        self.machine.game.balls_in_play += 1
        playfield.add_ball()
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.game.balls_in_play)

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()

        # ball directly enters the hold (again)
        self.machine.switch_controller.process_switch("s_ball_switch_hold1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, hold.balls)

        # playfield count goes to 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self._captured = 0

        # wait for eject confirm for hold
        self.advance_time_and_run(10)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        self.advance_time_and_run(100)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

    def test_auto_capacity(self):
        self.assertEqual(self.machine.ball_holds["hold_test3"].config['balls_to_hold'], 2)

    def test_enabled_state_in_placeholder(self):
        placeholder = self.machine.placeholder_manager.build_bool_template("device.ball_holds.hold_test.enabled")
        value, subscription = placeholder.evaluate_and_subscribe([])
        self.assertFalse(value)
        self.assertFalse(subscription.done())

        self.mock_event("should_post_when_enabled")
        self.mock_event("should_post_when_disabled")
        self.mock_event("should_not_post_when_enabled")
        self.mock_event("should_not_post_when_disabled")
        hold = self.machine.ball_holds['hold_test']

        hold.enable()
        self.assertTrue(hold.enabled)
        self.post_event("test_event_when_enabled")
        self.assertEventCalled("should_post_when_enabled")
        self.assertEventNotCalled("should_not_post_when_enabled")
        self.assertTrue(subscription.done())
        value, subscription = placeholder.evaluate_and_subscribe([])
        self.assertTrue(value)

        hold.disable()
        self.assertFalse(hold.enabled)
        self.post_event("test_event_when_disabled")
        self.assertEventCalled("should_post_when_disabled")
        self.assertEventNotCalled("should_not_post_when_disabled")
        self.assertTrue(subscription.done())
        value, subscription = placeholder.evaluate_and_subscribe([])
        self.assertFalse(value)

        # we need to chancel the subscription or it will crash later
        subscription.cancel()

class TestBallHoldSmart(MpfTestCase):

    def get_config_file(self):
        return 'test_ball_holds.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_holds/'

    def get_platform(self):
        return 'smart_virtual'

    def test_ball_end(self):
        self.machine.config['game']['balls_per_game'] = self.machine.placeholder_manager.build_int_template(3)
        # add an initial ball to trough
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(0, self.machine.playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # press start
        self.hit_and_release_switch("s_start")

        # wait until ball is on pf
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.game.player.ball)
        self.assertEqual(1, self.machine.playfield.balls)

        # hold one ball
        self.machine.ball_holds["hold_test"].enable()
        self.hit_switch_and_run("s_ball_switch_hold1", 1)

        # still held
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.ball_holds["hold_test"].balls_held)
        self.assertEqual(0, self.machine.playfield.balls)

        # request another ball
        self.machine.playfield.add_ball(1)
        self.advance_time_and_run()
        self.hit_and_release_switch('s_playfield_active')
        self.advance_time_and_run()
        self.assertEqual(1, self.machine.playfield.balls)

        # drain ball on pf
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["test_trough"])
        self.advance_time_and_run(1)

        # machine waits for the hold
        self.advance_time_and_run(10)
        self.assertEqual(1, self.machine.game.player.ball)

        # release the hold
        self.post_event('release_test')
        self.assertEqual(0, self.machine.ball_holds["hold_test"].balls_held)
        self.assertEqual(1, self.machine.playfield.balls)

        # once the ball drains move on to the ball 2
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["test_trough"])
        self.advance_time_and_run(1)
        self.assertEqual(2, self.machine.game.player.ball)

        # next ball
        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(1)

        # drain again
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["test_trough"])
        self.advance_time_and_run(1)

        # there should be no hold this time
        self.assertEqual(3, self.machine.game.player.ball)



"""Physical segment displays."""
from collections import namedtuple
from operator import attrgetter
from typing import List, Optional

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.placeholder_manager import TextTemplate
from mpf.core.system_wide_device import SystemWideDevice
from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplayPlatformInterface     # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.core.platform import SegmentDisplayPlatform    # pylint: disable-msg=cyclic-import,unused-import; # noqa

TextStack = namedtuple("TextStack", ["text", "priority", "key"])


@DeviceMonitor("text")
class SegmentDisplay(SystemWideDevice):

    """A physical segment display in a pinball machine."""

    config_section = 'segment_displays'
    collection = 'segment_displays'
    class_label = 'segment_display'

    def __init__(self, machine, name: str) -> None:
        """Initialise segment display device."""
        super().__init__(machine, name)
        self.hw_display = None                  # type: Optional[SegmentDisplayPlatformInterface]
        self.platform = None                    # type: Optional[SegmentDisplayPlatform]
        self._text_stack = []                   # type: List[TextStack]
        self._current_placeholder = None        # type: Optional[TextTemplate]
        self.text = ""                          # type: Optional[str]
        self.flashing = FlashingType.NO_FLASH   # type: FlashingType

    async def _initialize(self):
        await super()._initialize()
        """Initialise display."""
        # load platform
        self.platform = self.machine.get_platform_sections('segment_displays', self.config['platform'])
        self.platform.assert_has_feature("segment_displays")

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Segment Display must have a number.", 1)

        # configure hardware
        try:
            self.hw_display = await self.platform.configure_segment_display(self.config['number'],
                                                                            self.config['platform_settings'])
        except AssertionError as e:
            raise AssertionError("Error in platform while configuring segment display {}. "
                                 "See error above.".format(self.name)) from e

    def add_text(self, text: str, priority: int = 0, key: str = None) -> None:
        """Add text to display stack.

        This will replace texts with the same key.
        """
        # remove old text in case it has the same key
        self._text_stack[:] = [x for x in self._text_stack if x.key != key]
        # add new text
        self._text_stack.append(TextStack(text, priority, key))
        self._update_stack()

    def set_flashing(self, flashing: FlashingType):
        """Enable/Disable flashing."""
        self.flashing = flashing
        # invalidate text to force an update
        self.text = None
        self._update_display()

    def remove_text_by_key(self, key: str):
        """Remove entry from text stack."""
        self._text_stack[:] = [x for x in self._text_stack if x.key != key]
        self._update_stack()

    def _update_stack(self) -> None:
        """Sort stack and show top entry on display."""
        # do nothing if stack is emtpy. set display empty
        assert self.hw_display is not None
        if not self._text_stack:
            self.hw_display.set_text("", flashing=FlashingType.NO_FLASH)
            if self._current_placeholder:
                self.text = ""
                self._current_placeholder = None
            return

        # sort stack by priority
        self._text_stack.sort(key=attrgetter("priority"), reverse=True)
        # get top entry
        top_entry = self._text_stack[0]

        self._current_placeholder = TextTemplate(self.machine, top_entry.text)
        self._update_display()

    def _update_display(self, *args, **kwargs) -> None:
        """Update display to current text."""
        del args
        del kwargs
        assert self.hw_display is not None
        if not self._current_placeholder:
            new_text = ""
        else:
            new_text, future = self._current_placeholder.evaluate_and_subscribe({})
            future.add_done_callback(self._update_display)

        # set text to display if it changed
        if new_text != self.text:
            self.text = new_text
            self.hw_display.set_text(self.text, flashing=self.flashing)



from unittest.mock import MagicMock

from mpf.platforms.interfaces.driver_platform_interface import PulseSettings

from mpf.core.platform import SwitchSettings, DriverSettings

from mpf.tests.MpfTestCase import MpfTestCase


class TestAutofire(MpfTestCase):
    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/autofire/'

    def test_hw_rule_pulse(self):
        self.machine.default_platform.set_pulse_on_hit_rule = MagicMock()
        self.machine.autofires["ac_test"].enable()

        self.machine.default_platform.set_pulse_on_hit_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=True)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.autofires["ac_test"].disable()

        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=True))

    def test_hw_rule_overwrites(self):
        self.machine.default_platform.set_pulse_on_hit_rule = MagicMock()
        self.machine.autofires["ac_test_overwrites"].enable()

        self.machine.default_platform.set_pulse_on_hit_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test"].hw_switch, invert=False, debounce=True),
            DriverSettings(hw_driver=self.machine.coils["c_test"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=False)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.autofires["ac_test_overwrites"].disable()

        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test"].hw_switch, invert=False, debounce=True),
            DriverSettings(hw_driver=self.machine.coils["c_test"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=False))

        self.machine.default_platform.set_pulse_on_hit_rule = MagicMock()
        self.machine.autofires["ac_test_overwrites2"].enable()

        self.machine.default_platform.set_pulse_on_hit_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test_debounce_on"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test_recycle_off"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10), hold_settings=None, recycle=True)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.autofires["ac_test_overwrites2"].disable()

        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test_debounce_on"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test_recycle_off"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10), hold_settings=None, recycle=True))

    def test_hw_rule_coil_and_switch_defaults(self):
        self.machine.default_platform.set_pulse_on_hit_rule = MagicMock()
        self.machine.autofires["ac_test_defaults"].enable()

        self.machine.default_platform.set_pulse_on_hit_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test_debounce_on"].hw_switch, invert=False, debounce=True),
            DriverSettings(hw_driver=self.machine.coils["c_test_recycle_off"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10), hold_settings=None, recycle=False)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.autofires["ac_test_defaults"].disable()

        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test_debounce_on"].hw_switch, invert=False, debounce=True),
            DriverSettings(hw_driver=self.machine.coils["c_test_recycle_off"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10), hold_settings=None, recycle=False))

    def test_hw_rule_pulse_inverted_switch(self):
        self.machine.default_platform.set_pulse_on_hit_rule = MagicMock()
        self.machine.autofires["ac_test_inverted"].enable()

        self.machine.default_platform.set_pulse_on_hit_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test_nc"].hw_switch, invert=True, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test2"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=True)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.autofires["ac_test_inverted"].disable()

        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test_nc"].hw_switch, invert=True, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test2"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=True))

    def test_hw_rule_pulse_inverted_autofire(self):
        self.machine.default_platform.set_pulse_on_hit_rule = MagicMock()
        self.machine.autofires["ac_test_inverted2"].enable()

        self.machine.default_platform.set_pulse_on_hit_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test"].hw_switch, invert=True, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test2"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=True)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.autofires["ac_test_inverted2"].disable()

        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_test"].hw_switch, invert=True, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_test2"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=23), hold_settings=None, recycle=True))

    def test_disabled(self):
        """Verify that a disabled autofire coil doesn't post 'playfield_active'."""
        self.mock_event("playfield_active")
        self.machine_run()

        self.hit_and_release_switch("s_test_disabled")
        self.machine_run()
        self.assertEventNotCalled("playfield_active")

        self.machine.autofires["ac_test_disabled"].enable()
        self.hit_and_release_switch("s_test_disabled")
        self.machine_run()
        self.assertEventCalled("playfield_active", times=1)

        self.machine.autofires["ac_test_disabled"].disable()
        self.hit_and_release_switch("s_test_disabled")
        self.machine_run()
        self.assertEventCalled("playfield_active", times=1)

    def test_timeout(self):
        self.machine.autofires["ac_test_timeout"].enable()
        self.machine_run()

        # 9 hits are ok
        for _ in range(9):
            self.hit_and_release_switch("s_test")
            self.machine_run()

        self.assertTrue(self.machine.autofires["ac_test_timeout"]._enabled)

        # 10th hit should disable it
        self.hit_and_release_switch("s_test")
        self.machine_run()
        self.assertFalse(self.machine.autofires["ac_test_timeout"]._enabled)

        # reenable after 500ms
        self.advance_time_and_run(.6)
        self.assertTrue(self.machine.autofires["ac_test_timeout"]._enabled)

        # exire the older hits
        self.advance_time_and_run(1)

        # 9 hits are ok
        for _ in range(9):
            self.hit_and_release_switch("s_test")
            self.machine_run()

        self.assertTrue(self.machine.autofires["ac_test_timeout"]._enabled)

        # wait 1s
        self.advance_time_and_run(1)

        # another 9 hits are ok
        for _ in range(9):
            self.hit_and_release_switch("s_test")
            self.machine_run()

        self.assertTrue(self.machine.autofires["ac_test_timeout"]._enabled)

        # 10th hit should disable it
        self.hit_and_release_switch("s_test")
        self.machine_run()
        self.assertFalse(self.machine.autofires["ac_test_timeout"]._enabled)

        self.advance_time_and_run(.2)

        # disable manually while disabled by too many hits
        self.machine.autofires["ac_test_timeout"].disable()
        self.assertFalse(self.machine.autofires["ac_test_timeout"]._enabled)

        # should not reenable
        self.advance_time_and_run(.4)
        self.assertFalse(self.machine.autofires["ac_test_timeout"]._enabled)



"""Contains the parent class for custom code classes."""
from mpf.core.delays import DelayManager
from mpf.core.logging import LogMixin


class CustomCode(LogMixin):

    """Baseclass for custom code in a machine."""

    __slots__ = ["machine", "name", "delay"]

    def __init__(self, machine, name):
        """Initialise custom code."""
        super().__init__()
        self.machine = machine
        self.name = name

        self.configure_logging('CustomCode.' + name, 'basic', 'full')
        self.delay = DelayManager(self.machine)
        self.on_load()

    def __repr__(self):
        """Return string representation."""
        return '<Scriptlet.{}>'.format(self.name)

    def on_load(self):
        """Automatically called when this custom code class loads.

        It's the intention that the programmer will overwrite this method
        in his custom code.
        """



"""Contains the Attract class which is the attract mode in a pinball machine."""

from mpf.core.mode import Mode


class Attract(Mode):

    """Default mode running in a machine when a game is not in progress.

    The attract mode's main job is to watch for the start button to be pressed,
    to post the requests to start games, and to move the machine flow to the
    next mode if the request to start game comes back as approved.
    """

    __slots__ = ["start_button_pressed_time", "start_hold_time", "start_buttons_held"]

    def __init__(self, *args, **kwargs):
        """Initialise mode."""
        super().__init__(*args, **kwargs)

        self.start_button_pressed_time = 0.0
        self.start_hold_time = 0.0
        self.start_buttons_held = list()

    def mode_start(self, **kwargs):
        """Start the attract mode."""
        # register switch handlers for the start button press so we can
        # capture long presses

        # add these to the switch_handlers set so they'll be removed

        for switch in self.machine.switches.items_tagged(
                self.machine.config['game']['start_game_switch_tag']):
            self.switch_handlers.append(
                self.machine.switch_controller.add_switch_handler_obj(
                    switch, self.start_button_pressed, 1))
            self.switch_handlers.append(
                self.machine.switch_controller.add_switch_handler_obj(
                    switch, self.start_button_released, 0))

        if self.machine.config['game']['start_game_event']:
            self.add_mode_event_handler(self.machine.config['game']['start_game_event'], self.start_button_released)

        if hasattr(self.machine, 'ball_devices'):
            self.machine.ball_controller.collect_balls()

        # trigger ball search if we are missing balls and have at least one ball
        if 0 < self.machine.ball_controller.num_balls_known < self.machine.config['machine']['balls_installed']:
            for playfield in self.machine.playfields.values():
                playfield.ball_search.enable()
                playfield.ball_search.start()

    def start_button_pressed(self):
        """Handle start button press.

        Called when the a switch tagged with *start* is activated.

        Note that in MPF, the game start process is initiated when the start
        button is *released*, so when the button is first pressed, MPF just
        records the time stamp. This allows the total time the start button
        was pressed to be note, so that, for example, different types of games
        can be started based on long-presses of the start button.

        """
        self.start_button_pressed_time = self.machine.clock.get_time()

    def start_button_released(self, **kwargs):
        """Handle start button release.

        Called when the a switch tagged with *start* is deactivated.

        Since this is the Attract mode, this method posts a boolean event
        called *request_to_start_game*. If that event comes back True, this
        method calls :meth:`result_of_start_request`.
        """
        del kwargs
        self.start_hold_time = self.machine.clock.get_time() - self.start_button_pressed_time
        self.start_buttons_held = list()

        for switch in self.machine.switches.items_tagged('player'):
            if self.machine.switch_controller.is_active(switch):
                self.start_buttons_held.append(switch.name)

        self.machine.events.post_boolean('request_to_start_game',
                                         callback=self.result_of_start_request)
        '''event: request_to_start_game
        desc: This event is posted when to start a game. This is a boolean
        event. Any handler can return *False* and the game will not be
        started. Otherwise when this event is done, a new game is started.

        Posting this event is the only way to start a game in MPF, since many
        systems have to "approve" the start. (Are the balls in the right
        places, are there enough credits, etc.)
        '''

    def result_of_start_request(self, ev_result=True):
        """Handle the result of the start request.

        Called after the *request_to_start_game* event is posted.

        If `result` is True, this method posts the event
        *game_start*. If False, nothing happens, as the game start
        request was denied by some handler.

        Args:
        ----
            ev_result : Bool result of the boolean event
                *request_to_start_game.* If any registered event handler did not
                want the game to start, this will be False. Otherwise it's True.
        """
        if ev_result is False:
            self.debug_log("Game start was denied")
        else:  # else because we want to start on True *or* None
            self.debug_log("Let's start a game!!")
            self.machine.events.post('game_start',
                                     buttons=self.start_buttons_held,
                                     hold_time=self.start_hold_time)
            '''event: game_start
            desc: Starts game while bypassing the many systems which have to
            "approve" the start. (Are the balls in the right places, are there
            enough credits, etc.) Use of this method is not recommended but may
            be useful in testing code. Instead, use the *request_to_start_game*
            event.

            args:
            buttons: A list of switches tagged with *player* that were held in
            when the start button was released. This is used for "alternate"
            game starts (e.g. hold the right flipper and press start for
            tournament mode, etc.)

            hold_time: The time, in seconds, that the start button was held in
            to start the game. This can be used to start alternate games via a
            "long press" of the start button.
            '''



"""Contains the Combo Switch device class."""
from functools import partial

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor("state")
class ComboSwitch(SystemWideDevice, ModeDevice):

    """Combo Switch device."""

    config_section = 'combo_switches'
    collection = 'combo_switches'
    class_label = 'combo_switch'

    __slots__ = ["states", "_state", "_switches_1_active", "_switches_2_active", "delay", "_switch_handlers"]

    def __init__(self, machine, name):
        """Initialize Combo Switch."""
        super().__init__(machine, name)
        self.states = ['inactive', 'both', 'one']
        self._state = 'inactive'
        self._switches_1_active = False
        self._switches_2_active = False

        self.delay = DelayManager(self.machine)
        self._switch_handlers = []

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Validate and parse config."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)

        for state in self.states + ["switches_1", "switches_2"]:
            if not config['events_when_{}'.format(state)]:
                config['events_when_{}'.format(state)] = [
                    "{}_{}".format(self.name, state)]

        return config

    async def device_added_system_wide(self):
        """Add event handlers."""
        await super().device_added_system_wide()
        self._add_switch_handlers()

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Add event handlers."""
        self._add_switch_handlers()

    def _add_switch_handlers(self):
        if self.config['tag_1']:
            for tag in self.config['tag_1']:
                for switch in self.machine.switches.items_tagged(tag):
                    self.config['switches_1'].add(switch)

        if self.config['tag_2']:
            for tag in self.config['tag_2']:
                for switch in self.machine.switches.items_tagged(tag):
                    self.config['switches_2'].add(switch)

        self._register_switch_handlers()

    @property
    def state(self):
        """Return current state."""
        return self._state

    @property
    def can_exist_outside_of_game(self):
        """Return true if this device can exist outside of a game."""
        return True

    def device_removed_from_mode(self, mode):
        """Mode ended.

        Args:
        ----
            mode: mode which stopped
        """
        del mode

        self._remove_switch_handlers()
        self._kill_delays()

    def _register_switch_handlers(self):
        for switch in self.config['switches_1']:
            self._switch_handlers.append(switch.add_handler(self._switch_1_went_active, state=1, return_info=True))
            self._switch_handlers.append(switch.add_handler(self._switch_1_went_inactive, state=0, return_info=True))

        for switch in self.config['switches_2']:
            self._switch_handlers.append(switch.add_handler(self._switch_2_went_active, state=1, return_info=True))
            self._switch_handlers.append(switch.add_handler(self._switch_2_went_inactive, state=0, return_info=True))

    def _remove_switch_handlers(self):
        self.machine.switch_controller.remove_switch_handler_by_keys(self._switch_handlers)
        self._switch_handlers = []

    def _kill_delays(self):
        self.delay.clear()

    def _switch_1_went_active(self, switch_name, **kwargs):
        del kwargs
        self.debug_log('A switch from switches_1 just went active')
        self.delay.remove('switch_1_inactive')

        if self._switches_1_active:
            return

        if not self.config['hold_time']:
            self._activate_switches_1(switch_name)
        else:
            self.delay.add_if_doesnt_exist(self.config['hold_time'],
                                           partial(self._activate_switches_1, switch_name),
                                           'switch_1_active')

    def _switch_2_went_active(self, switch_name, **kwargs):
        del kwargs
        self.debug_log('A switch from switches_2 just went active')
        self.delay.remove('switch_2_inactive')

        if self._switches_2_active:
            return

        if not self.config['hold_time']:
            self._activate_switches_2(switch_name)
        else:
            self.delay.add_if_doesnt_exist(self.config['hold_time'],
                                           partial(self._activate_switches_2, switch_name),
                                           'switch_2_active')

    def _switch_1_went_inactive(self, switch_name, **kwargs):
        del kwargs
        self.debug_log('A switch from switches_1 just went inactive')
        for switch in self.config['switches_1']:
            if switch.state:
                # at least one switch is still active
                return

        self.delay.remove('switch_1_active')

        if not self.config['release_time']:
            self._release_switches_1(switch_name)
        else:
            self.delay.add_if_doesnt_exist(self.config['release_time'],
                                           partial(self._release_switches_1, switch_name),
                                           'switch_1_inactive')

    def _switch_2_went_inactive(self, switch_name, **kwargs):
        del kwargs
        self.debug_log('A switch from switches_2 just went inactive')
        for switch in self.config['switches_2']:
            if switch.state:
                # at least one switch is still active
                return

        self.delay.remove('switch_2_active')

        if not self.config['release_time']:
            self._release_switches_2(switch_name)
        else:
            self.delay.add_if_doesnt_exist(self.config['release_time'],
                                           partial(self._release_switches_2, switch_name),
                                           'switch_2_inactive')

    def _activate_switches_1(self, switch_name):
        self.debug_log('Switches_1 has passed the hold time and is now '
                       'active')
        self._switches_1_active = self.machine.clock.get_time()
        self.delay.remove("switch_2_only")

        if self._switches_2_active:
            if (self.config['max_offset_time'] >= 0 and
                    (self._switches_1_active - self._switches_2_active >
                        self.config['max_offset_time'])):

                self.debug_log("Switches_2 is active, but the "
                               "max_offset_time=%s which is largest than when "
                               "a Switches_2 switch was first activated, so "
                               "the state will not switch to 'both'",
                               self.config['max_offset_time'])

                return

            self._switch_state('both', group=1, switch=switch_name)
        elif self.config['max_offset_time'] >= 0:
            self.delay.add_if_doesnt_exist(self.config['max_offset_time'] * 1000, self._post_only_one_active_event,
                                           "switch_1_only", number=1)

    def _activate_switches_2(self, switch_name):
        self.debug_log('Switches_2 has passed the hold time and is now '
                       'active')
        self._switches_2_active = self.machine.clock.get_time()
        self.delay.remove("switch_1_only")

        if self._switches_1_active:
            if (self.config['max_offset_time'] >= 0 and
                    (self._switches_2_active - self._switches_1_active >
                        self.config['max_offset_time'])):
                self.debug_log("Switches_2 is active, but the "
                               "max_offset_time=%s which is largest than when "
                               "a Switches_2 switch was first activated, so "
                               "the state will not switch to 'both'",
                               self.config['max_offset_time'])
                return

            self._switch_state('both', group=2, switch=switch_name)
        elif self.config['max_offset_time'] >= 0:
            self.delay.add_if_doesnt_exist(self.config['max_offset_time'] * 1000, self._post_only_one_active_event,
                                           "switch_2_only", number=2)

    def _post_only_one_active_event(self, number):
        for event in self.config['events_when_switches_{}'.format(number)]:
            self.machine.events.post(event)

    def _release_switches_1(self, switch_name):
        self.debug_log('Switches_1 has passed the release time and is now '
                       'releases')
        self._switches_1_active = None
        if self._switches_2_active and self._state == 'both':
            self._switch_state('one', group=1, switch=switch_name)
        elif self._state == 'one':
            self._switch_state('inactive', group=1, switch=switch_name)

    def _release_switches_2(self, switch_name):
        self.debug_log('Switches_2 has passed the release time and is now '
                       'releases')
        self._switches_2_active = None
        if self._switches_1_active and self._state == 'both':
            self._switch_state('one', group=2, switch=switch_name)
        elif self._state == 'one':
            self._switch_state('inactive', group=2, switch=switch_name)

    def _switch_state(self, state, group, switch):
        """Post events for current step."""
        if state not in self.states:
            raise ValueError("Received invalid state: {}".format(state))

        if state == self.state:
            return

        self._state = state
        self.debug_log("New State: %s", state)

        for event in self.config['events_when_{}'.format(state)]:
            self.machine.events.post(event, triggering_group=group, triggering_switch=switch)
            '''event: (name)_one
            config_attribute: events_when_one

            desc: Combo switch (name) changed to state one.

            Either switch 1 or switch 2 has been released for at
            least the ``release_time:`` but the other switch is still active.
            '''

            '''event: (name)_both
            config_attribute: events_when_both

            desc: Combo switch (name) changed to state both.

            A switch from group 1 and group 2 are both active at the
            same time, having been pressed within the ``max_offset_time:`` and
            being active for at least the ``hold_time:``.
            '''

            '''event: (name)_inactive
            config_attribute: events_when_inactive

            desc: Combo switch (name) changed to state inactive.

            Both switches are inactive.
            '''

            '''event: (name)_switches_1
            config_attribute: events_when_switches_1

            desc: Combo switch (name) changed to state switches_1.

            Only switches_1 is active. max_offset_time has passed and this hit
            cannot become both later on. Only emited when ``max_offset_time:``
            is defined.
            '''

            '''event: (name)_switches_2
            config_attribute: events_when_switches_2

            desc: Combo switch (name) changed to state switches_2.

            Only switches_2 is active. max_offset_time has passed and this hit
            cannot become both later on. Only emited when ``max_offset_time:``
            is defined.
            '''



"""Config player for text on segment displays."""
from mpf.core.delays import DelayManager

from mpf.config_players.device_config_player import DeviceConfigPlayer
from mpf.platforms.interfaces.segment_display_platform_interface import FlashingType

MYPY = False
if MYPY:   # pragma: no cover
    from typing import Dict     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.segment_display import SegmentDisplay  # pylint: disable-msg=cyclic-import,unused-import


class SegmentDisplayPlayer(DeviceConfigPlayer):

    """Generates texts on segment displays."""

    config_file_section = 'segment_display_player'
    show_section = 'segment_displays'
    machine_collection_name = 'segment_displays'

    __slots__ = ["delay"]

    def __init__(self, machine):
        """Initialise SegmentDisplayPlayer."""
        super().__init__(machine)
        self.delay = DelayManager(self.machine)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Show text on display."""
        del kwargs
        instance_dict = self._get_instance_dict(context)    # type: Dict[str, SegmentDisplay]
        full_context = self._get_full_context(context)

        for display, s in settings.items():
            action = s['action']
            if display not in instance_dict:
                instance_dict[display] = {}

            key = full_context + "." + display.name

            if s['key']:
                key += s['key']

            if action == "add":
                # add text
                display.add_text(s['text'], priority + s['priority'], key)

                if s['expire']:
                    instance_dict[display][key] = self.delay.add(
                        s['expire'], self._remove,
                        instance_dict=instance_dict,
                        key=key, display=display)
                else:
                    instance_dict[display][key] = True
            elif action == "remove":
                self._remove(instance_dict=instance_dict,
                             key=key, display=display)
            elif action == "flash":
                display.set_flashing(FlashingType.FLASH_ALL)
            elif action == "flash_match":
                display.set_flashing(FlashingType.FLASH_MATCH)
            elif action == "no_flash":
                display.set_flashing(FlashingType.NO_FLASH)
            else:
                raise AssertionError("Invalid action {}".format(action))

    def _remove(self, instance_dict, key, display):
        if key in instance_dict[display]:
            display.remove_text_by_key(key)
            if instance_dict[display][key] is not True:
                self.delay.remove(instance_dict[display][key])
            del instance_dict[display][key]

    def clear_context(self, context):
        """Remove all texts."""
        instance_dict = self._get_instance_dict(context)
        for display, keys in instance_dict.items():
            for key in dict(keys).keys():
                self._remove(instance_dict=instance_dict,
                             key=key, display=display)

        self._reset_instance_dict(context)

    def get_express_config(self, value):
        """Parse express config."""
        return dict(action="add", text=value)



"""Contains the Game class which is the Machine Mode that actually runs and manages the game in a pinball machine.

Note that in the Mission Pinball Framework, a distinction is made between a
*game* and a *machine*. A *game* refers to a game in progress, whereas a
*machine* is the physical pinball machine.
"""
from functools import partial
import asyncio

from mpf.core.events import QueuedEvent
from mpf.core.async_mode import AsyncMode
from mpf.core.player import Player


# pylint: disable-msg=too-many-instance-attributes
class Game(AsyncMode):

    """Base mode that runs an active game on a pinball machine.

    The game mode is responsible for creating players, starting and ending
    balls, rotating to the next player, etc.
    """

    __slots__ = ["_balls_in_play", "player_list", "slam_tilted", "tilted", "ending", "player", "num_players",
                 "_stopping_modes", "_stopping_queue", "_end_ball_event", "_at_least_one_player_event",
                 "balls_per_game", "max_players"]

    def __init__(self, *args, **kwargs):
        """Initialise game."""
        super().__init__(*args, **kwargs)
        self._balls_in_play = 0
        self.player_list = list()
        self.machine.game = None
        self.slam_tilted = False
        self.tilted = False
        self.ending = False
        self.player = None
        self.num_players = None
        self._stopping_modes = []
        self._stopping_queue = None
        self._end_ball_event = None  # type: asyncio.Event
        self._at_least_one_player_event = None  # type: asyncio.Event
        self.balls_per_game = None
        self.max_players = None

        self.machine.events.add_handler('mode_{}_stopping'.format(self.name), self._stop_game_modes)

    async def _run(self):
        """Run the basic game loop."""
        # Init the game (game start process)
        # Initialize member variables
        self.player = None
        self.player_list = list()
        self.machine.game = self
        self.slam_tilted = False
        self.tilted = False
        self.ending = False
        self.num_players = 0
        self._balls_in_play = 0
        self._stopping_modes = []
        self._stopping_queue = None
        self._end_ball_event = asyncio.Event()
        self._end_ball_event.clear()
        self._at_least_one_player_event = asyncio.Event()
        self._at_least_one_player_event.clear()
        self.balls_per_game = self.machine.config['game']['balls_per_game'].evaluate([])

        # Add add player switch handler
        if self.machine.config['game']['add_player_switch_tag']:
            self.add_mode_event_handler(
                self.machine.config['mpf']['switch_tag_event'].replace(
                    '%', self.machine.config['game']['add_player_switch_tag']),
                self.request_player_add)

        if self.machine.config['game']['add_player_event']:
            self.add_mode_event_handler(self.machine.config['game']['add_player_event'], self.request_player_add)

        self.max_players = self.machine.config['game']['max_players'].evaluate({})

        self.add_mode_event_handler(self.machine.config['game']['end_ball_event'], self.event_end_ball)
        self.add_mode_event_handler(self.machine.config['game']['end_game_event'], self.event_end_game)

        await self._start_game()

        # Game loop
        while not self.ending:
            await self._start_player_turn()
            # run the ball
            await self._run_ball()

            # run any extra balls
            while self.player.extra_balls and not self.slam_tilted:
                await self._award_extra_ball()

            await self._end_player_turn()

            if self.slam_tilted or self.player.ball >= self.balls_per_game and self.player.number == self.num_players:
                self.ending = True
            else:
                await self._rotate_players()

        await self._end_game()

    async def _run_ball(self, is_extra_ball=False):
        self._end_ball_event.clear()
        await self._start_ball(is_extra_ball)
        # Wait for end ball event to be set
        await self._end_ball_event.wait()
        await self._end_ball()

    async def _start_game(self):
        """Start a new game."""
        self.debug_log("Game start")
        await self.machine.events.post_async('game_will_start')
        '''event: game_will_start
        desc: The game is about to start. This event is posted just before
        :doc:`game_starting`.'''

        await self.machine.events.post_queue_async('game_starting', game=self)
        '''event: game_starting
        desc: A game is in the process of starting. This is a queue event, and
        the game won't actually start until the queue is cleared.

        args:
        game: A reference to the game mode object.
        '''

        # Sometimes game_starting handlers will add players, so we only
        # have to add one here if there aren't any players yet.
        if self.player_list:
            self._at_least_one_player_event.set()
        else:
            self._at_least_one_player_event.clear()
            self.request_player_add()

        # Wait for player to be added before game can start
        # TODO: Add timeout to wait
        await self._at_least_one_player_event.wait()

        await self.machine.events.post_async('game_started')
        '''event: game_started
        desc: A new game has started.'''

        self.debug_log("Game started")

    def ball_ending(self):
        """Handle ball ending.

        DEPRECATED in v0.50. Use ``end_ball()`` instead.
        """
        # TODO: Remove this function as it has been deprecated and replaced
        self.warning_log("game.ball_ending() function has been deprecated. "
                         "Please use game.end_ball() instead.")
        self.end_ball()

    def event_end_game(self, **kwargs):
        """Event handler to end the current game."""
        del kwargs
        self.end_game()

    def event_end_ball(self, **kwargs):
        """Event handler to end the current ball."""
        del kwargs
        self.end_ball()

    def end_ball(self):
        """Set an event flag that will end the current ball."""
        self._end_ball_event.set()

    async def _end_ball(self):
        """Start the ball ending process.

        This method posts the queue event *ball_ending*, giving other modules
        an opportunity to finish up whatever they need to do before the ball
        ends. Once all the registered handlers for that event have finished,
        this method calls :meth:`ball_ended`.

        Currently this method also disables the autofire_coils and flippers,
        though that's temporary as we'll move those into config file options.
        """
        # remove the handlers that were looking for ball drain since they'll
        # be re-added on next ball start
        self.machine.events.remove_handler(self.ball_drained)

        # todo should clean up the above since they are removed from the
        # active list of handlers but not the registered_handlers list.
        # It doesn't really matter since the game ending can just remove them
        # all, but technically it's not clean.

        self._balls_in_play = 0

        self.debug_log("Entering Game._end_ball()")

        await self.machine.events.post_async('ball_will_end')
        '''event: ball_will_end
        desc: The ball is about to end. This event is posted just before
        :doc:`ball_ending`.'''

        await self.machine.events.post_queue_async('ball_ending')
        '''event: ball_ending
        desc: The ball is ending. This is a queue event and the ball won't
        actually end until the queue is cleared.

        This event is posted just after :doc:`ball_will_end`'''

        await self.machine.events.post_async('ball_ended')
        '''event: ball_ended
        desc: The ball has ended.

        Note that this does not necessarily mean that the next player's turn
        will start, as this player may have an extra ball which means they'll
        shoot again.'''

        self.debug_log("Ball has ended")

    @property
    def balls_in_play(self) -> int:
        """Property which holds the current number of balls in play.

        Note that the number of balls in play is not necessarily the same as
        the number of balls that are active on the playfield. (For example,
        a ball could be held in a device while a show is playing, etc.)

        You can set this property to change it, or get it's value.

        If you set this value to 0, the ball ending process will be started.

        """
        return self._balls_in_play

    @balls_in_play.setter
    def balls_in_play(self, value: int):
        """Set balls in play."""
        prev_balls_in_play = self._balls_in_play

        if value > self.machine.ball_controller.num_balls_known:
            self.warning_log("Balls in play have been capped by num_balls_known.")
            self._balls_in_play = self.machine.ball_controller.num_balls_known

        elif value < 0:
            self._balls_in_play = 0

        else:
            self._balls_in_play = value

        self.debug_log("Balls in Play change. New value: %s, (Previous: %s)",
                       self._balls_in_play, prev_balls_in_play)

        if self._balls_in_play > 0:
            self.machine.events.post('balls_in_play',
                                     balls=self._balls_in_play)
            '''event: balls_in_play
            desc: The number of balls in play has just changed, and there is at
            least 1 ball in play.

            Note that the number of balls in play is not necessarily the same
            as the number of balls loose on the playfield. For example, if the
            player shoots a lock and is watching a cut scene, there is still
            one ball in play even though there are no balls on the playfield.

            args:
            balls: The number of ball(s) in play.'''

        if prev_balls_in_play and not self._balls_in_play:
            # Trigger end ball process by setting event
            self._end_ball_event.set()

    def _stop_game_modes(self, queue: QueuedEvent, **kwargs):
        """Stop all game modes and wait until they stopped."""
        del kwargs
        self._stopping_modes = []
        for mode in self.machine.modes.values():
            if mode.is_game_mode and mode.active:
                self._stopping_modes.append(mode)
                mode.stop(callback=partial(self._game_mode_stopped, mode=mode))

        if self._stopping_modes:
            queue.wait()
            self._stopping_queue = queue

    @property
    def is_game_mode(self):
        """Return false.

        We are the game and not a mode within the game.
        """
        return False

    def _game_mode_stopped(self, mode):
        """Mark game mode stopped and clear the wait on stop if this was the last one."""
        self._stopping_modes.remove(mode)
        if not self._stopping_modes:
            self._stopping_queue.clear()
            self._stopping_queue = None

    def mode_stop(self, **kwargs):
        """Stop mode."""
        del kwargs

        for mode in self.machine.modes.values():
            if mode.active and mode.is_game_mode:
                raise AssertionError("Mode {} is not supposed to run outside of game."
                                     .format(mode.name))
        self.machine.game = None

    async def _start_ball(self, is_extra_ball=False):
        """Perform ball start procedure.

        Note this method is called for each ball that starts, even if it's
        after a Shoot Again scenario for the same player.

        Posts a queue event called *ball_starting*, giving other modules the
        opportunity to do things before the ball actually starts. Once that
        event is clear, this method calls :meth:`ball_started`.
        """
        event_args = {
            "player": self.player.number,
            "ball": self.player.ball,
            "balls_remaining": self.balls_per_game - self.player.ball,
            "is_extra_ball": is_extra_ball}

        self.debug_log("***************************************************")
        self.debug_log("****************** BALL STARTING ******************")
        self.debug_log("**                                               **")
        self.debug_log("**    Player: {}    Ball: {}   Score: {}".format(self.player.number,
                                                                         self.player.ball,
                                                                         self.player.score
                                                                         ).ljust(49) + '**')
        self.debug_log("**                                               **")
        self.debug_log("***************************************************")
        self.debug_log("***************************************************")

        await self.machine.events.post_async('ball_will_start', **event_args)
        '''event: ball_will_start
        desc: The ball is about to start. This event is posted just before
        :doc:`ball_starting`.
        args:
        ball: The ball number
        balls_remaining: The number of balls left in the game (not including this one)
        is_extra_ball: True if this ball is an extra ball (default False)
        player: The player number'''

        await self.machine.events.post_queue_async('ball_starting', **event_args)
        '''event: ball_starting
        desc: A ball is starting. This is a queue event, so the ball won't
        actually start until the queue is cleared.
        args:
        ball: The ball number
        balls_remaining: The number of balls left in the game (not including this one)
        is_extra_ball: True if this ball is an extra ball (default False)
        player: The player number'''

        # register handlers to watch for ball drain and live ball removed
        self.add_mode_event_handler('ball_drain', self.ball_drained)

        self.balls_in_play = 1

        self.debug_log("ball_started for Ball %s", self.player.ball)

        await self.machine.events.post_async('ball_started', **event_args)
        '''event: ball_started
        desc: A new ball has started.
        args:
        ball: The ball number
        balls_remaining: The number of balls left in the game (not including this one)
        is_extra_ball: True if this ball is an extra ball (default False)
        player: The player number'''

        if self.num_players == 1:
            await self.machine.events.post_async('single_player_ball_started')
            '''event: single_player_ball_started
            desc: A new ball has started, and this is a single player game.'''
        else:
            await self.machine.events.post_async('multi_player_ball_started')
            '''event: multi_player_ball_started
            desc: A new ball has started, and this is a multiplayer game.'''
            await self.machine.events.post_async(
                'player_{}_ball_started'.format(self.player.number))
            '''event player_(number)_ball_started
            desc: A new ball has started, and this is a multiplayer game.
            The player number is the (number) in the event that's posted.'''

        if not hasattr(self.machine, "playfield") or not self.machine.playfield:
            raise AssertionError("The game did not define default playfield. Did you add tags: default to one of your "
                                 "playfield?")

        self.machine.playfield.add_ball(player_controlled=True)

    def ball_drained(self, balls=0, **kwargs):
        """One or more balls has drained.

        Drained balls will be subtracted from the number of balls in play.

        Args:
        ----
            balls: The number of balls that just drained.

        Returns a dictionary: {balls: *number of balls drained*}.
        """
        del kwargs
        self.debug_log("Entering Game.ball_drained()")

        if balls:
            self.debug_log("Processing %s newly-drained ball(s)", balls)
            self.balls_in_play -= balls

        return {'balls': balls}

    async def _end_game(self):
        self.debug_log("Entering Game._end_game()")

        await self.machine.events.post_async('game_will_end')
        '''event: game_will_end
        desc: The game is about to end. This event is posted just before
        :doc:`game_ending`.'''

        await self.machine.events.post_queue_async('game_ending')
        '''event: game_ending
        desc: The game is in the process of ending. This is a queue event, and
        the game won't actually end until the queue is cleared.'''

        # set playerX_score variables
        if self.player_list:
            for player in self.player_list:
                self.machine.variables.configure_machine_var(name='player{}_score'.format(player.number), persist=True)
                self.machine.variables.set_machine_var(
                    name='player{}_score'.format(player.number),
                    value=player.score)
                '''machine_var: player(x)_score

                desc: Holds the numeric value of a player's score from the last
                game. The "x" is the player number, so this actual machine
                variable is ``player1_score`` or ``player2_score``.

                Since these are machine variables, they are maintained even after
                a game is over. Therefore you can use these machine variables in
                your attract mode display show to show the scores of the last game
                that was played.

                These machine variables are updated at the end of the game,
                and they persist on disk so they are restored the next time
                MPF starts up.
                '''

            # remove all other vars
            for i in range(len(self.player_list) + 1, self.max_players + 1):
                self.machine.variables.remove_machine_var('player{}_score'.format(i))

        await self.machine.events.post_async('game_ended')
        '''event: game_ended
        desc: The game has ended.'''

    def game_ending(self):
        """Handle game ending.

        DEPRECATED in v0.50. Use ``end_game()`` instead.
        """
        # TODO: Remove this function as it has been deprecated and replaced
        self.warning_log("game.game_ending() function has been deprecated. "
                         "Please use game.end_game() instead.")
        self.end_game()

    def end_game(self):
        """End the current game.

        This triggers the game end manually.
        """
        self.ending = True
        self.end_ball()

    def _game_ending_completed(self, **kwargs):
        del kwargs

        if self.player:

            self.machine.events.post('player_turn_will_end',
                                     player=self.player,
                                     number=self.player.number)
            self.machine.events.post_queue('player_turn_ending',
                                           player=self.player,
                                           number=self.player.number,
                                           callback=self._player_turn_at_game_ending_completed)
        else:
            self.machine.events.post('game_ended')

    def _player_turn_at_game_ending_completed(self, **kwargs):
        del kwargs
        self.machine.events.post('player_turn_ended',
                                 player=self.player,
                                 number=self.player.number)

        self.machine.events.post('game_ended')

    async def _award_extra_ball(self):
        """Award an extra ball to the player.

        Player gained an extra ball during play. Same player should shoot again.
        """
        self.debug_log("Awarded extra ball to Player %s. Shoot Again", self.player.index + 1)
        self.player.extra_balls -= 1
        await self._run_ball(is_extra_ball=True)

    def request_player_add(self, **kwargs):
        """Request to add a player to an active game.

        This method contains the logic to verify whether it's ok to add a
        player. For example, the game must be on Ball 1 and the current
        number of players must be less than the max number allowed.

        Assuming this method believes it's ok to add a player, it posts the
        boolean event *player_add_request* to give other modules the
        opportunity to deny it. For example, a credits module might deny the
        request if there are not enough credits in the machine.

        If *player_add_request* comes back True, the event
        *player_added* is posted with a reference to the new player
        object as a *player* kwarg.

        """
        del kwargs
        self.debug_log("Received request to add player.")

        # There area few things we have to check first. If this all passes,
        # then we'll raise the event to ask other modules if it's ok to add a
        # player

        if self.ending:
            self.debug_log("Game is ending. Cannot add player.")
            return False

        if len(self.player_list) >= self.max_players:
            self.debug_log("Game is at max players. Cannot add another.")
            return False

        if self.player and self.player.ball > 1:  # todo config setting
            self.debug_log("Current ball is after Ball 1. Cannot add player.")
            return False

        self.machine.events.post_boolean('player_add_request',
                                         callback=self._player_add_request_complete)
        '''event: player_add_request
        desc: Posted to request that an additional player be added to this
        game. Any registered handler can deny the player add request by
        returning *False* to this event.
        '''
        return True

    def _player_add_request_complete(self, ev_result=True, **kwargs) -> bool:
        """Handle result of player_add_request callback.

        Callback from our request player add event. Don't call it directly.
        """
        del kwargs
        if ev_result is False:
            self.debug_log("Request to add player has been denied.")
            return False

        new_player_number = len(self.player_list) + 1

        self.machine.events.post('player_will_add', number=new_player_number)
        '''event: player_will_add
        desc: A new player will be added to this game. This event is sent immediately
        prior to the player_adding event.

        args:
        number: The new player number that will be added
        '''

        # Actually create the new player object
        player = Player(self.machine, len(self.player_list))

        self.player_list.append(player)
        self.num_players = len(self.player_list)

        self.machine.events.post_queue('player_adding',
                                       player=player,
                                       number=player.number,
                                       callback=partial(self._player_adding_complete,
                                                        player=player))
        '''event: player_adding
        desc: A new player is in the process of being added to this game. This is a queue
        event, and the player won't actually be finished adding until the queue is cleared.

        args:
        player: The player object for the player being added
        number: The player number
        '''
        return True

    def _player_adding_complete(self, player, **kwargs):
        """Handle result of the player_adding queue event.

        The add player process can now finish.
        """
        del kwargs
        self.machine.events.post('player_added',
                                 player=player,
                                 num=player.number,
                                 callback=self._player_added)
        '''event: player_added

        desc: A new player was just added to this game

        args:

        player: A reference to the instance of the Player() object.

        num: The number of the player that was just added. (e.g. Player 1 will
        have *num=1*, Player 4 will have *num=4*, etc.)
        '''

        # set player if there is none
        if not self.player:
            self.player = player

        # At least one player has been added to the current game, set event
        self._at_least_one_player_event.set()

        self.info_log("Player added successfully. Total players: %s", self.num_players)

        if self.num_players == 2:
            self.machine.events.post('multiplayer_game')
            '''event: multiplayer_game
             desc: A second player has just been added to this game, meaning
             this is now a multiplayer game.

             This event is typically used to switch the score display from the
             single player layout to the multiplayer layout.
             '''

        return True

    @staticmethod
    def _player_added(player, num):
        # Now that the player_added event has been posted, enable player
        # variable events and send all initial values
        del num
        player.enable_events(True, True)

    async def _start_player_turn(self):
        """Start the players turn.

        Note this method is only called when a different player's turn is up.
        So if the same player shoots again due to an extra ball, this method
        is not called again.
        """
        # If we get a request to start a turn but we haven't done a rotate to
        # set the first player, do that now.
        if not self.player:
            await self._rotate_players()

        await self.machine.events.post_async('player_turn_will_start',
                                             player=self.player,
                                             number=self.player.number)
        '''event: player_turn_will_start
        desc: A new player's turn will start. This event is only posted before the
        start of a new player's turn. If that player gets an extra ball and
        shoots again, this event is not posted a second time.

        args:
        player: The player object whose turn is starting.
        number: The player number
        '''

        await self.machine.events.post_queue_async('player_turn_starting',
                                                   player=self.player,
                                                   number=self.player.number)
        '''event: player_turn_starting
        desc: The player's turn is in the process of starting. This is a queue
        event, and the player's turn won't actually start until the queue is cleared.

        args:
        player: The player object whose turn is starting.
        number: The player number
        '''

        self.player.ball += 1
        '''player_var: ball

        desc: The ball number for this player. If a player gets an extra ball,
        this number won't change when they start the extra ball.
        '''

        await self.machine.events.post_async('player_turn_started',
                                             player=self.player,
                                             number=self.player.number)
        '''event: player_turn_started
        desc: A new player's turn started. This event is only posted after the
        start of a new player's turn. If that player gets an extra ball and
        shoots again, this event is not posted a second time.

        args:
        player: The player object whose turn is starting.
        number: The player number
        '''

    async def _end_player_turn(self):
        """End the current player's turn."""
        if not self.player:
            return

        await self.machine.events.post_async('player_turn_will_end',
                                             player=self.player,
                                             number=self.player.number)
        '''event: player_turn_will_end
        desc: The player's turn is about to end. This event is only posted when this
        player's turn is totally over. If the player gets an extra ball and
        shoots again, this event is not posted until after all their extra
        balls and it's no longer their turn.

        args:
        player: The player object whose turn is over.
        number: The player number
        '''

        await self.machine.events.post_queue_async('player_turn_ending',
                                                   player=self.player,
                                                   number=self.player.number)
        '''event: player_turn_ending
        desc: The current player's turn is ending. This is a queue event, and
        the player's turn won't actually end until the queue is cleared.

        args:
        player: The player object whose turn is ending.
        number: The player number
        '''

        await self.machine.events.post_async('player_turn_ended',
                                             player=self.player,
                                             number=self.player.number)
        '''event: player_turn_ended
        desc: The current player's turn has ended. This event is only posted when
        this player's turn is totally over. If the player gets an extra ball and
        shoots again, this event is not posted until after all their extra balls
        and it's no longer their turn.

        args:
        player: The player object whose turn is ending.
        number: The player number
        '''

    async def _rotate_players(self):
        """Rotate the game to the next player.

        This method is called after a player's turn is over, so it's even used
        in single-player games between balls.

        All it does really is set :attr:`player` to the next player's number.
        """
        # todo  do cool stuff in the future to change order, etc.

        if self.player and self.player.number < self.num_players:
            self.player = self.player_list[self.player.number]
            # Note the above line is kind of confusing but it works because
            # the current player number is always 1 more than the index.
            # i.e. "Player 1" has an index of 0, etc. So using the current
            # player number as the next player's index works out.
        else:
            # no current player, grab the first one
            self.player = self.player_list[0]

        self.debug_log("Player rotate: Now up is Player %s",
                       self.player.number)

# todo player events should come next, including tracking inc/dec, other values



"""Test sequence shots."""
from mpf.tests.MpfTestCase import MpfTestCase


class TestShots(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/sequence_shot/'

    def test_simple_sequence(self):
        self.mock_event("sequence1_hit")
        self.post_event("event1")
        self.machine_run()
        self.post_event("event2")
        self.machine_run()
        self.assertEventNotCalled("sequence1_hit")
        self.post_event("event3")
        self.machine_run()
        self.assertEventCalled("sequence1_hit")
        self.mock_event("sequence1_hit")
        self.post_event("event1")
        self.machine_run()
        self.post_event("event2")
        self.machine_run()
        self.assertEventNotCalled("sequence1_hit")
        self.post_event("event3")
        self.machine_run()
        self.assertEventCalled("sequence1_hit")

    def test_delay(self):
        self.mock_event("sequence1_hit")
        self.post_event("delay1")
        # no success after delay
        self.machine_run()
        self.post_event("event1")
        self.machine_run()
        self.post_event("event2")
        self.machine_run()
        self.post_event("event3")
        self.machine_run()
        self.assertEventNotCalled("sequence1_hit")

        # works after delay
        self.advance_time_and_run(1.5)
        self.post_event("event1")
        self.machine_run()
        self.post_event("event2")
        self.machine_run()
        self.assertEventNotCalled("sequence1_hit")
        self.post_event("event3")
        self.machine_run()
        self.assertEventCalled("sequence1_hit")

    def test_delay_switch(self):
        self.mock_event("sequence2_hit")
        self.hit_and_release_switch("seq2_delay")
        self.machine_run()
        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()
        self.hit_and_release_switch("seq2_3")
        self.machine_run()
        self.assertEventNotCalled("sequence2_hit")

        # works after delay
        self.advance_time_and_run(1.5)
        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()
        self.assertEventNotCalled("sequence2_hit")
        self.hit_and_release_switch("seq2_3")
        self.machine_run()
        self.assertEventCalled("sequence2_hit")

    def test_simple_switch_sequence(self):
        self.mock_event("sequence2_hit")
        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()
        self.assertEventNotCalled("sequence2_hit")
        self.hit_and_release_switch("seq2_3")
        self.machine_run()
        self.assertEventCalled("sequence2_hit")

    def test_cancel(self):
        self.mock_event("sequence1_hit")
        self.post_event("event1")
        self.machine_run()
        self.post_event("event2")
        self.machine_run()
        self.assertEventNotCalled("sequence1_hit")
        self.post_event("cancel")
        self.machine_run()
        self.post_event("event3")
        self.machine_run()
        self.assertEventNotCalled("sequence1_hit")

    def test_cancel_switch(self):
        self.mock_event("sequence2_hit")
        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()
        self.assertEventNotCalled("sequence2_hit")
        self.hit_and_release_switch("seq2_cancel")
        self.hit_and_release_switch("seq2_3")
        self.machine_run()
        self.assertEventNotCalled("sequence2_hit")

    def test_single_step_sequence(self):
        self.mock_event("sequence3_hit")
        self.post_event("event3_1")
        self.machine_run()
        self.assertEventCalled("sequence3_hit")

    def test_sequence_with_duplicates(self):
        self.assertEqual(1, len(self.machine.events.registered_handlers["event_1"]))
        self.assertEqual(1, len(self.machine.events.registered_handlers["event_2"]))
        self.assertEqual(1, len(self.machine.events.registered_handlers["event_3"]))

    def test_interleaved_sequences(self):
        """"Two balls pass through the sequence."""
        self.mock_event("sequence1_hit")

        self.post_event("event1")
        self.advance_time_and_run(.2)
        self.post_event("event1")
        self.advance_time_and_run(.1)

        self.post_event("event2")
        self.advance_time_and_run(.2)
        self.post_event("event2")
        self.advance_time_and_run(.1)

        self.assertEventNotCalled("sequence1_hit")
        self.post_event("event3")
        self.advance_time_and_run(.2)
        self.assertEventCalled("sequence1_hit")
        self.mock_event("sequence1_hit")
        self.post_event("event3")
        self.assertEventCalled("sequence1_hit")

    def test_sequence_timeout(self):
        """Ball rolls up a ramp and back down."""
        self.mock_event("sequence1_hit")
        self.mock_event("sequence1_timeout")
        self.post_event("event1")
        self.advance_time_and_run(.2)
        self.post_event("event2")
        self.advance_time_and_run(.1)
        self.post_event("event2")
        self.advance_time_and_run(.2)
        self.post_event("event1")
        self.advance_time_and_run(.1)
        self.assertEventNotCalled("sequence1_hit")
        self.assertEventNotCalled("sequence1_timeout")
        self.advance_time_and_run(2.5)
        self.assertEventCalled("sequence1_timeout", times=1)
        self.advance_time_and_run(.5)
        # second timeout from interleaved sequence. can we prevent this?
        self.assertEventCalled("sequence1_timeout", times=2)

    def test_mode_seqence(self):
        """"Test sequence in mode."""
        self.mock_event("sequence_mode_event_hit")
        self.mock_event("sequence_mode_switch_hit")

        self.post_event("event1")
        self.advance_time_and_run(.2)
        self.post_event("event2")
        self.advance_time_and_run(.2)

        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()

        self.assertEventNotCalled("sequence_mode_event_hit")
        self.assertEventNotCalled("sequence_mode_switch_hit")

        self.start_mode("mode1")
        self.assertEventNotCalled("sequence_mode_event_hit")
        self.assertEventNotCalled("sequence_mode_switch_hit")

        self.post_event("event1")
        self.advance_time_and_run(.2)
        self.post_event("event2")
        self.advance_time_and_run(.2)

        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()

        self.assertEventCalled("sequence_mode_event_hit")
        self.assertEventCalled("sequence_mode_switch_hit")
        self.mock_event("sequence_mode_event_hit")
        self.mock_event("sequence_mode_switch_hit")

        self.stop_mode("mode1")

        self.post_event("event1")
        self.advance_time_and_run(.2)
        self.post_event("event2")
        self.advance_time_and_run(.2)

        self.hit_and_release_switch("seq2_1")
        self.machine_run()
        self.hit_and_release_switch("seq2_2")
        self.machine_run()

        self.assertEventNotCalled("sequence_mode_event_hit")
        self.assertEventNotCalled("sequence_mode_switch_hit")

    def test_single_switch_sequence(self):
        self.mock_event("sequence4_hit")
        self.hit_and_release_switch("seq4_delay")
        self.machine_run()
        self.hit_and_release_switch("seq4_1")
        self.machine_run()
        self.assertEventNotCalled("sequence4_hit")

        # works after delay
        self.advance_time_and_run(1.5)
        self.hit_and_release_switch("seq4_1")
        self.machine_run()
        self.assertEventCalled("sequence4_hit")




"""Mode which allows the player to select another mode to run."""
from mpf.core.utility_functions import Util

from mpf.core.mode import Mode


class Carousel(Mode):

    """Mode which allows the player to select another mode to run."""

    __slots__ = ["_all_items", "_items", "_select_item_events", "_next_item_events",
                 "_previous_item_events", "_highlighted_item_index", "_done",
                 "_block_events", "_release_events", "_is_blocking"]

    def __init__(self, *args, **kwargs):
        """Initialise carousel mode."""
        self._all_items = None
        self._items = None
        self._select_item_events = None
        self._next_item_events = None
        self._previous_item_events = None
        self._highlighted_item_index = None
        self._block_events = None
        self._release_events = None
        self._is_blocking = None
        self._done = None
        super().__init__(*args, **kwargs)

    def mode_init(self):
        """Initialise mode and read all settings from config."""
        super().mode_init()
        mode_settings = self.config.get("mode_settings", [])
        self._all_items = []
        for item in Util.string_to_event_list(mode_settings.get("selectable_items", "")):
            placeholder = self.machine.placeholder_manager.parse_conditional_template(item)
            # Only add a placeholder if there's a condition, otherwise just the string
            self._all_items.append(placeholder if placeholder.condition else item)
        self._select_item_events = Util.string_to_event_list(mode_settings.get("select_item_events", ""))
        self._next_item_events = Util.string_to_event_list(mode_settings.get("next_item_events", ""))
        self._previous_item_events = Util.string_to_event_list(mode_settings.get("previous_item_events", ""))
        self._highlighted_item_index = 0
        self._block_events = Util.string_to_event_list(mode_settings.get("block_events", ""))
        self._release_events = Util.string_to_event_list(mode_settings.get("release_events", ""))

        if not self._all_items:
            raise AssertionError("Specify at least one item to select from")

    def mode_start(self, **kwargs):
        """Start mode and let the player select."""
        self._items = []
        for item in self._all_items:
            # All strings go in, but only conditional templates if they evaluate true
            if isinstance(item, str):
                self._items.append(item)
            elif not item.condition or item.condition.evaluate({}):
                self._items.append(item.name)
        if not self._items:
            self.machine.events.post("{}_items_empty".format(self.name))
            '''event (carousel_name)_items_empty
                desc: A carousel's items are all conditional and all evaluated false.
                    If this event is posted, the carousel mode will not be started.
                '''
            self.stop()
            return

        super().mode_start(**kwargs)
        self._done = False
        self._is_blocking = False

        self._register_handlers(self._next_item_events, self._next_item)
        self._register_handlers(self._previous_item_events, self._previous_item)
        self._register_handlers(self._select_item_events, self._select_item)

        self._highlighted_item_index = 0
        self._update_highlighted_item(None)

        # If set to block next/prev on flipper cancel, set those event handlers
        if self._block_events:
            # This rudimentary implementation will block on any block_event
            # and release on any release_event. If future requirements need to
            # track *which* block_event blocked and *only* release on the
            # corresponding release_event, additional work will be needed.
            for event in self._block_events:
                self.add_mode_event_handler(event, self._block_enable, priority=100)
            for event in self._release_events:
                self.add_mode_event_handler(event, self._block_disable, priority=100)

    def _register_handlers(self, events, handler):
        for event in events:
            self.add_mode_event_handler(event, handler)

    def _get_highlighted_item(self):
        return self._get_available_items()[self._highlighted_item_index]

    def _update_highlighted_item(self, direction):
        self.debug_log("Highlighted item: " + self._get_highlighted_item())

        self.machine.events.post("{}_{}_highlighted".format(self.name, self._get_highlighted_item()),
                                 direction=direction)
        '''event (carousel_name)_(item)_highlighted
            desc: Player highlighted an item in a carousel. Mostly used to play shows or trigger slides.
            args:
               direction: The direction the carousel is moving. Either forwards or backwards. None on mode start.
            '''

    def _get_available_items(self):
        # Return the default items
        return self._items

    def _next_item(self, **kwargs):
        del kwargs
        if self._done or self._is_blocking:
            return
        self._highlighted_item_index += 1
        if self._highlighted_item_index >= len(self._get_available_items()):
            self._highlighted_item_index = 0
        self._update_highlighted_item("forwards")

    def _previous_item(self, **kwargs):
        del kwargs
        if self._done or self._is_blocking:
            return
        self._highlighted_item_index -= 1
        if self._highlighted_item_index < 0:
            self._highlighted_item_index = len(self._get_available_items()) - 1

        self._update_highlighted_item("backwards")

    def _select_item(self, **kwargs):
        del kwargs
        if self._done:
            return
        self.debug_log("Selected mode: " + str(self._get_highlighted_item()))
        self._done = True

        self.machine.events.post("{}_{}_selected".format(self.name, self._get_highlighted_item()))
        '''event (carousel_name)_(item)_selected
            desc: Player selected an item in a carousel. Can be used to trigger modes. '''
        self.machine.events.post("{}_item_selected".format(self.name))
        '''event (carousel_name)_item_selected
            desc: Player selected any item in a carousel. Used to stop mode. '''

    def _block_enable(self, **kwargs):
        del kwargs
        self._is_blocking = True

    def _block_disable(self, **kwargs):
        del kwargs
        self._is_blocking = False
